# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for REST models."""

from typing import Any
import unittest
from unittest import mock
import langfun.core as lf
from langfun.core.llms import rest
import pyglove as pg
import requests


def mock_requests_post(url: str, json: dict[str, Any], **kwargs):
  del url, kwargs
  response = requests.Response()
  response.status_code = 200
  response._content = pg.to_json_str({
      'content': [(
          f'hello with temperature={json.get("temperature")}, '
          f'top_k={json.get("top_k")}, '
          f'top_p={json.get("top_p")}, '
          f'max_tokens={json.get("max_tokens")}, '
          f'stop={json.get("stop_sequences")}.'
      )],
  }).encode()
  return response


def mock_requests_post_exception(error):
  def _mock_requests(url: str, json: dict[str, Any], **kwargs):
    del url, json, kwargs
    raise error
  return _mock_requests


def mock_requests_post_error(status_code, error_type, error_message):
  def _mock_requests(url: str, json: dict[str, Any], **kwargs):
    del url, json, kwargs
    response = requests.Response()
    response.status_code = status_code
    response._content = pg.to_json_str(
        {
            'error': {
                'type': error_type,
                'message': error_message,
            }
        }
    ).encode()
    return response

  return _mock_requests


class RestTest(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self._lm = rest.REST(
        api_endpoint='https://fake-api.com',
        request=lambda x, o: dict(
            model='test-model',
            prompt=x.text,
            temperature=0.0,
            top_k=0.1,
            top_p=0.2,
            stop_sequences=['\n'],
            max_tokens=4096,
        ),
        result=lambda x: lf.LMSamplingResult(
            [lf.LMSample(c) for c in x['content']]),
        headers=dict(api_key='fake_key'),
    )

  def test_call(self):
    with mock.patch('requests.Session.post') as mock_request:
      mock_request.side_effect = mock_requests_post
      self.assertEqual(self._lm.model_id, 'unknown')
      response = self._lm(
          'hello', temperature=0.0, top_k=0.1, top_p=0.2, stop=['\n'])
      self.assertEqual(
          response.text,
          (
              'hello with temperature=0.0, top_k=0.1, top_p=0.2, '
              "max_tokens=4096, stop=['\\n']."
          ),
      )
      self.assertIsInstance(response.usage, lf.UsageNotAvailable)

  def test_call_errors(self):
    for status_code, error_type, error_message in [
        (429, 'rate_limit', 'Rate limit exceeded.'),
        (529, 'service_unavailable', 'Service unavailable.'),
        (500, 'bad_request', 'Bad request.'),
    ]:
      with mock.patch('requests.Session.post') as mock_mm_request:
        mock_mm_request.side_effect = mock_requests_post_error(
            status_code, error_type, error_message
        )
        with self.assertRaisesRegex(
            Exception, f'.*{status_code}: .*{error_message}'
        ):
          self._lm('hello', max_attempts=1)

    for error, expected_lm_error_cls, expected_lm_error_msg in [
        (
            requests.exceptions.Timeout('Timeout.'),
            lf.TemporaryLMError,
            'Timeout.',
        ),
        (
            requests.exceptions.ReadTimeout('Read timeout.'),
            lf.TemporaryLMError,
            'Read timeout.',
        ),
        (
            requests.exceptions.ConnectTimeout('Connect timeout.'),
            lf.TemporaryLMError,
            'Connect timeout.',
        ),
        (
            TimeoutError('Timeout error.'),
            lf.TemporaryLMError,
            'Timeout error.',
        ),
        (
            requests.exceptions.ConnectionError('REJECTED_CLIENT_THROTTLED'),
            lf.TemporaryLMError,
            'REJECTED_CLIENT_THROTTLED',
        ),
        (
            requests.exceptions.ConnectionError('Connection error.'),
            lf.LMError,
            'Connection error.',
        ),
        (
            ConnectionError('Connection error.'),
            lf.LMError,
            'Connection error.',
        )
    ]:
      with mock.patch('requests.Session.post') as mock_post:
        mock_post.side_effect = mock_requests_post_exception(error)
        with self.assertRaisesRegex(
            expected_lm_error_cls, expected_lm_error_msg
        ):
          self._lm._sample_single(lf.UserMessage('hello'))


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for Gemini API."""

from typing import Any
import unittest
from unittest import mock

import langfun.core as lf
from langfun.core.llms import gemini
import pyglove as pg
import requests


example_image = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)


def mock_requests_post(url: str, json: dict[str, Any], **kwargs):
  del url, kwargs
  c = pg.Dict(json['generationConfig'])
  parts = []
  if system_instruction := json.get('systemInstruction'):
    parts.extend([p['text'] for p in system_instruction.get('parts', [])])

  # Add text from the main contents.
  for c_item in json.get('contents', []):
    for p in c_item.get('parts', []):
      parts.append(p['text'])
  content = '\n'.join(parts)
  response = requests.Response()
  response.status_code = 200
  response._content = pg.to_json_str({
      'candidates': [
          {
              'content': {
                  'parts': [
                      {
                          'text': (
                              f'This is a response to {content} with '
                              f'temperature={c.temperature}, '
                              f'top_p={c.topP}, '
                              f'top_k={c.topK}, '
                              f'max_tokens={c.maxOutputTokens}, '
                              f'stop={"".join(c.stopSequences)}.'
                          ),
                      },
                      {
                          'text': 'This is the thought.',
                          'thought': True,
                      }
                  ],
              },
          },
      ],
      'usageMetadata': {
          'promptTokenCount': 3,
          'candidatesTokenCount': 4,
      }
  }).encode()
  return response


class GeminiTest(unittest.TestCase):
  """Tests for Vertex model with REST API."""

  def test_dir(self):
    self.assertIn('gemini-1.5-pro', gemini.Gemini.dir())

  def test_estimate_cost(self):
    model = gemini.Gemini('gemini-1.5-pro', api_endpoint='')
    self.assertEqual(
        model.estimate_cost(
            lf.LMSamplingUsage(
                prompt_tokens=100_000,
                completion_tokens=1000,
                total_tokens=101_000,
            )
        ),
        0.13
    )
    # Prompt length greater than 128K.
    self.assertEqual(
        model.estimate_cost(
            lf.LMSamplingUsage(
                prompt_tokens=200_000,
                completion_tokens=1000,
                total_tokens=201_000,
            )
        ),
        0.51
    )

  def test_generation_config(self):
    model = gemini.Gemini('gemini-1.5-pro', api_endpoint='')
    json_schema = {
        'type': 'object',
        'properties': {
            'name': {'type': 'string'},
        },
        'required': ['name'],
        'title': 'Person',
    }
    actual = model._generation_config(
        lf.UserMessage('hi', json_schema=json_schema),
        lf.LMSamplingOptions(
            temperature=2.0,
            top_p=1.0,
            top_k=20,
            max_tokens=1024,
            stop=['\n'],
        ),
    )
    self.assertEqual(
        actual,
        dict(
            candidateCount=1,
            temperature=2.0,
            topP=1.0,
            topK=20,
            maxOutputTokens=1024,
            stopSequences=['\n'],
            responseLogprobs=False,
            logprobs=None,
            seed=None,
            responseMimeType='application/json',
            responseSchema={
                'type': 'object',
                'properties': {
                    'name': {'type': 'string'}
                },
                'required': ['name'],
                'title': 'Person',
            }
        ),
    )

    # Add test for thinkingConfig.
    actual = model._generation_config(
        lf.UserMessage('hi'),
        lf.LMSamplingOptions(
            max_thinking_tokens=100,
        ),
    )
    self.assertEqual(
        actual,
        dict(
            candidateCount=1,
            temperature=None,
            topP=None,
            topK=40,
            maxOutputTokens=None,
            stopSequences=None,
            responseLogprobs=False,
            logprobs=None,
            seed=None,
            thinkingConfig={'includeThoughts': True, 'thinkingBudget': 100},
        ),
    )

    with self.assertRaisesRegex(
        ValueError, '`json_schema` must be a dict, got'
    ):
      model._generation_config(
          lf.UserMessage('hi', json_schema='not a dict'),
          lf.LMSamplingOptions(),
      )

  def test_call_model(self):
    with mock.patch('requests.Session.post') as mock_generate:
      mock_generate.side_effect = mock_requests_post

      lm = gemini.Gemini('gemini-1.5-pro', api_endpoint='')
      r = lm(
          'hello',
          temperature=2.0,
          top_p=1.0,
          top_k=20,
          max_tokens=1024,
          stop='\n',
      )
      self.assertEqual(
          r.text,
          (
              'This is a response to hello with temperature=2.0, '
              'top_p=1.0, top_k=20, max_tokens=1024, stop=\n.'
          ),
      )
      self.assertEqual(r.metadata.thought, 'This is the thought.')
      self.assertEqual(r.metadata.usage.prompt_tokens, 3)
      self.assertEqual(r.metadata.usage.completion_tokens, 4)

  def test_call_model_with_context_limit_error(self):
    def mock_requests_post_error(*args, **kwargs):
      del args, kwargs
      response = requests.Response()
      response.status_code = 400
      response._content = b'exceeds the maximum number of tokens.'
      return response

    with mock.patch('requests.Session.post') as mock_generate:
      mock_generate.side_effect = mock_requests_post_error
      lm = gemini.Gemini('gemini-1.5-pro', api_endpoint='')
      with self.assertRaisesRegex(
          lf.ContextLimitError, 'exceeds the maximum number of tokens.'
      ):
        lm('hello')

  def test_call_model_with_system_message(self):
    with mock.patch('requests.Session.post') as mock_generate:
      mock_generate.side_effect = mock_requests_post

      lm = gemini.Gemini('gemini-1.5-pro', api_endpoint='')
      r = lm(
          lf.UserMessage('hello', system_message=lf.SystemMessage('system')),
          temperature=2.0,
          top_p=1.0,
          top_k=20,
          max_tokens=1024,
          stop='\n',
      )
      self.assertEqual(
          r.text,
          (
              'This is a response to system\nhello with temperature=2.0, '
              'top_p=1.0, top_k=20, max_tokens=1024, stop=\n.'
          ),
      )


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Anthropic API message conversion."""

import base64
from typing import Annotated, Any, Callable

import langfun.core as lf
from langfun.core import modalities as lf_modalities


class AnthropicMessageConverter(lf.MessageConverter):
  """Converter to Anthropic public API."""

  FORMAT_ID = 'anthropic'

  chunk_preprocessor: Annotated[
      Callable[[str | lf.Modality], Any] | None,
      (
          'Chunk preprocessor for Langfun chunk to Anthropic chunk conversion. '
          'It will be applied before each Langfun chunk is converted. '
          'If returns None, the chunk will be skipped.'
      )
  ] = None

  def to_value(self, message: lf.Message) -> dict[str, Any]:
    """Converts a Langfun message to Gemini API."""
    content = []
    for chunk in message.chunk():
      if self.chunk_preprocessor:
        chunk = self.chunk_preprocessor(chunk)
        if chunk is None:
          continue

      if isinstance(chunk, str):
        content.append({'type': 'text', 'text': chunk})
      elif isinstance(chunk, lf_modalities.Mime):
        if isinstance(chunk, lf_modalities.Image):
          content.append(
              dict(
                  type='image',
                  source=dict(
                      type='base64',
                      media_type=chunk.mime_type,
                      data=base64.b64encode(chunk.to_bytes()).decode(),
                  ),
              )
          )
        elif isinstance(chunk, lf_modalities.PDF):
          content.append(
              dict(
                  type='document',
                  source=dict(
                      type='base64',
                      media_type=chunk.mime_type,
                      data=base64.b64encode(chunk.to_bytes()).decode(),
                  ),
              )
          )
        else:
          raise NotImplementedError(
              f'Modality conversion not implemented: {chunk!r}'
          )
    return dict(role=self.get_role(message), content=content)

  def from_value(self, value: dict[str, Any]) -> lf.Message:
    """Returns a Langfun message from Anthropic message."""
    message_cls = self.get_message_cls(
        self._safe_read(value, 'role', default='assistant')
    )
    content = self._safe_read(value, 'content', default=[])
    assert isinstance(content, list)

    chunks = []
    thought_chunks = []
    for part in content:
      t = self._safe_read(part, 'type')
      if t == 'text':
        chunks.append(self._safe_read(part, 'text'))
      elif t == 'thinking':
        thought_chunks.append(self._safe_read(part, 'thinking'))
      elif t in ('image', 'document'):
        source = self._safe_read(part, 'source')
        chunks.append(
            lf_modalities.Mime.class_from_mime_type(
                self._safe_read(source, 'media_type')
            ).from_bytes(base64.b64decode(self._safe_read(source, 'data')))
        )
      else:
        raise ValueError(f'Unsupported content part: {part!r}.')
    message = message_cls.from_chunks(chunks)
    if thought_chunks:
      message.set('thought', message_cls.from_chunks(thought_chunks))
    return message


def _as_anthropic_format(
    self,
    chunk_preprocessor: Callable[[str | lf.Modality], Any] | None = None,
    **kwargs
) -> dict[str, Any]:
  """Returns an Anthropic format message."""
  return AnthropicMessageConverter(
      chunk_preprocessor=chunk_preprocessor, **kwargs
  ).to_value(self)


@classmethod
def _from_anthropic_format(
    cls,
    anthropic_message: dict[str, Any],
    **kwargs
) -> lf.Message:
  """Creates a Langfun message from the Anthropic format message."""
  del cls
  return AnthropicMessageConverter(**kwargs).from_value(anthropic_message)

# Set shortcut methods in lf.Message.
lf.Message.as_anthropic_format = _as_anthropic_format
lf.Message.from_anthropic_format = _from_anthropic_format



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Conversation."""

from typing import Annotated
import langfun.core as lf
import langfun.core.memories as lf_memories
from langfun.core.templates.completion import Completion


@lf.use_init_args(['preamble', 'input_message'])
class Conversation(Completion):
  """LM-based conversation."""

  prompt = """
      {%- if preamble -%}
      {{ preamble }}
      {%- endif%}
      {%if conversation_context -%}
      {{ conversation_context }}
      {%- endif%}
      {{ next_input }}
      """

  preamble: Annotated[
      lf.LangFunc | None,
      '(Optional) Preamble before beginning the conversation.',
  ] = None

  role: Annotated[
      str | None,
      '(Optional) User defined role for the AI response in the conversation.',
  ] = None

  conversation_context: Annotated[
      lf.LangFunc | None,
      (
          '(Optional) Conversation context, which could be history of existing '
          'chat messages.'
      ),
  ] = lf.LangFunc('{{ memory }}')

  memory: Annotated[
      lf.Memory, 'Memory to use in formulating the conversation context.'
  ] = lf_memories.ConversationHistory()

  next_input: Annotated[
      lf.LangFunc, '(Required) next input message for this chat.'
  ] = lf.LangFunc('User: {{ input_message }}')

  input_message: Annotated[
      lf.Message | str, '(Contextual) input message for this chat.'
  ] = lf.contextual()

  def _on_bound(self):
    super()._on_bound()
    self.reset()

  def __call__(self, **kwargs) -> lf.Message:
    """Sends a message to LM and receives a response."""

    # Clear (maybe) cached LM response from the last turn.
    self.clear_lm_response()

    with lf.context(**kwargs):
      # Call LM based on the prompt generated from `input_message`.
      lm_response = super().__call__()
      if self.role is not None:
        lm_response.rebind(
            sender=self.role, skip_notification=True, raise_on_no_change=False
        )

      # Add current turn to memory.
      self.add(self.input_message, lm_response)

    return lm_response

  def add(
      self, input_message: lf.Message | str, lm_response: lf.Message | str
  ) -> None:
    """Adds a turn."""
    self.memory.remember((input_message, lm_response))

  def reset(self) -> None:
    """Resets current conversation."""
    self.memory.reset()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for modality (e.g. Image, Video, etc.)."""

import abc
import functools
import hashlib
from typing import Any, ContextManager
from langfun.core import component
import pyglove as pg


_TLS_MODALITY_AS_REF = '__format_modality_as_ref__'


def format_modality_as_ref(enabled: bool = True) -> ContextManager[None]:
  """A context manager that formats modality objects as references."""
  return pg.object_utils.thread_local_value_scope(
      _TLS_MODALITY_AS_REF, enabled, False
  )


class Modality(component.Component, pg.views.HtmlTreeView.Extension):
  """Base class for multimodal object."""

  REF_START = '<<[['
  REF_END = ']]>>'

  def _on_bound(self):
    super()._on_bound()
    # Invalidate cached hash if modality member is changed.
    self.__dict__.pop('hash', None)

  def format(self, *args, **kwargs) -> str:
    if self.referred_name is None or not pg.object_utils.thread_local_get(
        _TLS_MODALITY_AS_REF, False
    ):
      return super().format(*args, **kwargs)
    return Modality.text_marker(self.referred_name)

  def __str_kwargs__(self) -> dict[str, Any]:
    # For modality objects, we don't want to use markdown format when they
    # are rendered as parts of the prompt.
    kwargs = super().__str_kwargs__()
    kwargs.pop('markdown', None)
    return kwargs

  @abc.abstractmethod
  def to_bytes(self) -> bytes:
    """Returns content in bytes."""

  @functools.cached_property
  def hash(self) -> str:
    """Returns a 8-byte MD5 hash as the identifier for this modality object."""
    return hashlib.md5(self.to_bytes()).hexdigest()[:8]

  @classmethod
  def text_marker(cls, var_name: str) -> str:
    """Returns a marker in the text for this object."""
    return Modality.REF_START + var_name + Modality.REF_END

  @property
  def referred_name(self) -> str | None:
    """Returns the referred name of this object in its template."""
    if not self.sym_path:
      return None
    # Strip the metadata prefix under message.
    path = str(self.sym_path)
    return path[9:] if path.startswith('metadata.') else path

  @classmethod
  def from_value(cls, value: pg.Symbolic) -> dict[str, 'Modality']:
    """Returns a dict of path to modality from a symbolic value."""
    modalities = {}
    def _visit(k, v, p):
      del k, p
      if isinstance(v, Modality):
        modalities[v.referred_name] = v
        return pg.TraverseAction.CONTINUE
      return pg.TraverseAction.ENTER

    pg.traverse(value, _visit)
    return modalities


class ModalityRef(pg.Object, pg.typing.CustomTyping):
  """References of modality objects in a symbolic tree.

  `ModalityRef` was introduced to placehold modality objects in a symbolic
  tree, to prevent message from being chunked in the middle of a Python
  structure.
  """

  name: str

  def custom_apply(
      self, path: pg.KeyPath, value_spec: pg.ValueSpec, *args, **kwargs
  ) -> tuple[bool, Any]:
    return (False, self)

  @classmethod
  def placehold(cls, value: pg.Symbolic) -> pg.Symbolic:
    """Returns a copy of value by replacing modality objects with refs.

    Args:
      value: A symbolic value.

    Returns:
      A copy of value with all child `Modality` objects replaced with
        `ModalityRef` objects.
    """

    def _placehold(k, v, p):
      del p
      if isinstance(v, Modality):
        return ModalityRef(name=value.sym_path + k)
      return v
    return value.clone().rebind(_placehold, raise_on_no_change=False)


class ModalityError(RuntimeError):  # pylint: disable=g-bad-exception-name
  """Exception raised when modality is not supported."""



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import metrics
import pyglove as pg

Example = example_lib.Example


class MatchTest(unittest.TestCase):

  def test_basic(self):
    m = metrics.Match()  # pylint: disable=invalid-name
    self.assertEqual(
        m.audit(Example(id=1, input=pg.Dict(groundtruth=1), output=1)),
        dict(match=True)
    )
    self.assertEqual(
        m.audit(Example(id=2, input=pg.Dict(groundtruth=1), output=2)),
        dict(mismatch=True)
    )
    self.assertEqual(
        m.audit(
            Example(
                id=3,
                input=pg.Dict(groundtruth=1),
                error=pg.symbolic.ErrorInfo(
                    tag='ValueError',
                    description='Bad input.',
                    stacktrace='...',
                )
            )
        ),
        dict(error='ValueError')
    )
    self.assertEqual(
        m.audit(
            Example(
                id=3,
                input=pg.Dict(groundtruth=1),
                error=pg.symbolic.ErrorInfo(
                    tag='MappingError.CodeError',
                    description='Bad input.',
                    stacktrace='...',
                )
            )
        ),
        dict(error='MappingError.CodeError')
    )
    self.assertEqual(m.matches, 0.25)
    self.assertEqual(m.mismatches, 0.25)
    self.assertEqual(m.oop_errors, 0.25)
    self.assertEqual(m.non_oop_errors, 0.25)

    self.assertEqual(m.values(), [
        m.matches,
        m.mismatches,
        m.oop_errors,
        m.non_oop_errors
    ])
    m.reset()
    self.assertEqual(len(m.matches.data_points), 0)
    self.assertEqual(len(m.mismatches.data_points), 0)
    self.assertEqual(len(m.oop_errors.data_points), 0)
    self.assertEqual(len(m.non_oop_errors.data_points), 0)

  def test_bad_case(self):
    m = metrics.Match()  # pylint: disable=invalid-name
    with self.assertRaisesRegex(ValueError, '`groundtruth` is not present'):
      m.audit(Example(id=1, input=pg.Dict(x=1), output=1))

  def test_custom_metadata(self):

    class MyMatch(metrics.Match):
      def match(self, example_input, output):
        return example_input.x == output, dict(x=example_input.x)

    m = MyMatch()  # pylint: disable=invalid-name
    self.assertEqual(
        m.audit(Example(id=1, input=pg.Dict(x=1), output=1)),
        dict(match=True, x=1)
    )
    self.assertEqual(m.matches, 1.0)

  def test_html_view(self):
    m = metrics.Match()  # pylint: disable=invalid-name
    m.audit(Example(id=1, input=pg.Dict(groundtruth=1), output=1))
    self.assertIn(
        '100.0%',
        m.to_html().content,
    )
    with pg.views.html.controls.HtmlControl.track_scripts() as scripts:
      m.audit(Example(id=2, input=pg.Dict(groundtruth=1), output=2))
      self.assertEqual(len(scripts), 12)


class ScoreTest(unittest.TestCase):

  def test_basic(self):

    class MyScore(metrics.Score):

      def score(self, example_input, output) -> float:
        return example_input.x * output

    m = MyScore()  # pylint: disable=invalid-name
    self.assertEqual(
        m.audit(Example(id=1, input=pg.Dict(x=1), output=1)),
        dict(score=1 * 1)
    )
    self.assertEqual(
        m.audit(Example(id=2, input=pg.Dict(x=2), output=2)),
        dict(score=2 * 2)
    )
    self.assertEqual(
        m.audit(
            Example(
                id=3,
                input=pg.Dict(x=1),
                error=pg.symbolic.ErrorInfo(
                    tag='ValueError',
                    description='Bad input.',
                    stacktrace='...',
                )
            )
        ),
        dict(error='ValueError')
    )
    self.assertEqual(
        m.audit(
            Example(
                id=3,
                input=pg.Dict(x=1),
                error=pg.symbolic.ErrorInfo(
                    tag='MappingError.CodeError',
                    description='Bad input.',
                    stacktrace='...',
                )
            )
        ),
        dict(error='MappingError.CodeError')
    )
    self.assertEqual(m.average_score, 2.5)
    self.assertEqual(m.oop_errors, 0.25)
    self.assertEqual(m.non_oop_errors, 0.25)

    self.assertEqual(m.values(), [
        m.average_score,
        m.oop_errors,
        m.non_oop_errors
    ])
    m.reset()
    self.assertEqual(len(m.average_score.data_points), 0)
    self.assertEqual(len(m.oop_errors.data_points), 0)
    self.assertEqual(len(m.non_oop_errors.data_points), 0)

  def test_custom_metadata(self):

    class MyScore(metrics.Score):

      def score(self, example_input, output):
        return example_input.x * output, dict(x=example_input.x)

    m = MyScore()  # pylint: disable=invalid-name
    self.assertEqual(
        m.audit(Example(id=1, input=pg.Dict(x=1), output=1)),
        dict(score=1 * 1, x=1)
    )
    self.assertEqual(m.average_score, 1.0)

  def test_html_view(self):

    class MyScore(metrics.Score):

      def score(self, example_input, output) -> float:
        return example_input.x * output

    m = MyScore()  # pylint: disable=invalid-name
    m.audit(Example(id=1, input=pg.Dict(x=1), output=2))
    self.assertIn(
        '2.000',
        m.to_html().content,
    )
    with pg.views.html.controls.HtmlControl.track_scripts() as scripts:
      m.audit(Example(id=2, input=pg.Dict(x=1), output=2))
      self.assertEqual(len(scripts), 9)


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for evaluation patching."""

import unittest
from langfun.core import llms as lf_llms
from langfun.core.eval import base
from langfun.core.eval import patching
import pyglove as pg


class PatchingCommonTest(unittest.TestCase):

  def test_patch_member(self):
    class A(pg.Object):
      x: int = 1

    class B(pg.Object):
      a: A

    b = B(A())
    pg.patch(b, [patching.patch_member(A, 'x', 2)])
    self.assertEqual(b, B(A(2)))

  def test_patch_args(self):
    s = base.Suite(
        [base.Evaluation(inputs=base.as_inputs([1]))],
        additional_args=dict(x=1, y=2),
    )
    pg.patch(s, [patching.patch_additional_args(x=3, z=4)])
    self.assertTrue(
        pg.eq(
            s,
            base.Suite(
                [
                    base.Evaluation(
                        inputs=base.as_inputs([1]),
                        additional_args=dict(x=3, y=2, z=4),
                    )
                ],
                additional_args=dict(x=3, y=2, z=4),
            ),
        )
    )

  def test_patch_lm(self):
    s = base.Suite(
        [base.Evaluation(inputs=base.as_inputs([1]))],
        lm=lf_llms.Gpt35Turbo(),
    )
    pg.patch(
        s, [patching.patch_lm(pg.oneof([lf_llms.Gpt35Turbo(), lf_llms.Gpt4()]))]
    )
    self.assertTrue(
        pg.eq(
            s,
            base.Suite(
                [
                    base.Evaluation(
                        inputs=base.as_inputs([1]),
                        lm=pg.oneof([lf_llms.Gpt35Turbo(), lf_llms.Gpt4()]),
                    )
                ],
                lm=pg.oneof([lf_llms.Gpt35Turbo(), lf_llms.Gpt4()]),
            ),
        )
    )

  def test_patch_parsing_lm(self):
    s = base.Suite(
        [base.Evaluation(inputs=base.as_inputs([1]))],
        lm=lf_llms.Gpt4(),
    )
    pg.patch(s, [patching.patch_parsing_lm(lf_llms.Gpt35Turbo())])
    self.assertTrue(
        pg.eq(
            s,
            base.Suite(
                [
                    base.Evaluation(
                        inputs=base.as_inputs([1]),
                        lm=lf_llms.Gpt4(),
                        parsing_lm=lf_llms.Gpt35Turbo(),
                    )
                ],
                # NOTE(daiyip): Suite does not have `parsing_lm` as one of its
                # variable keyword fields yet, so patching does not add to it.
                # This is okay since we only care about the leaf nodes.
                lm=lf_llms.Gpt4(),
            ),
        )
    )

  def test_patch_prompt(self):
    e = base.Evaluation(inputs=base.as_inputs([1]))
    pg.patch(e, [patching.patch_prompt('Q: {{example.question}}')])
    self.assertTrue(
        pg.eq(
            e,
            base.Evaluation(
                inputs=base.as_inputs([1]),
                prompt='Q: {{example.question}}',
            ),
        )
    )

  def test_patch_inputs(self):
    e = base.Evaluation(inputs=base.as_inputs([1]))
    pg.patch(e, [patching.patch_inputs(base.as_inputs([2]))])
    self.assertTrue(
        pg.eq(
            e,
            base.Evaluation(
                inputs=base.as_inputs([2]),
            ),
        )
    )

  def test_patch_schema_fn(self):
    @pg.functor()
    def int_schema():
      return int

    e = base.Evaluation(inputs=base.as_inputs([1]))
    pg.patch(e, [patching.patch_schema_fn(int_schema())])
    self.assertTrue(
        pg.eq(
            e,
            base.Evaluation(
                inputs=base.as_inputs([1]),
                schema_fn=int_schema(),
            ),
        )
    )


class StringPatcheTest(unittest.TestCase):

  def test_lm(self):
    target = pg.patch(
        base.Evaluation(inputs=base.as_inputs([1])),
        ['lm?haiku:gpt4', 'max_tokens?1024', 'temperature?0.7'],
    )
    self.assertEqual(
        target.lm,
        pg.oneof([
            lf_llms.Claude3Haiku(temperature=0.7, max_tokens=1024),
            lf_llms.Gpt4(temperature=0.7, max_tokens=1024),
        ]),
    )
    with self.assertRaisesRegex(ValueError, 'Unknown model name'):
      pg.patch(
          base.Evaluation(inputs=base.as_inputs([1])),
          ['lm?gpt2'],
      )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Text completion task."""

from typing import Annotated, Union
import langfun.core as lf


@lf.use_init_args(['prompt', 'lm_response'])
class Completion(lf.LangFunc):
  """LM-based text completion, this is base for all LM-based tasks.

  {{ prompt }}
  {% if response -%}
  {{ response }}
  {%- endif -%}
  """

  #
  # Template variables.
  #

  prompt: Annotated[
      lf.LangFunc,
      (
          'Prompt for completition.'
      )
  ]

  lm_response: Annotated[
      Union[None, str, lf.Message, lf.LangFunc],
      (
          '(Optional) LM rsesponse. This is the cached LM response when '
          '`cache_response` is set to True, or user provided response for '
          'training data generation.'
      ),
  ] = None

  # Task options.
  cache_response: bool = False

  @property
  def response(self) -> Union[None, str, lf.Message, lf.LangFunc]:
    """Formalized respone. Subclass could override."""
    return self.lm_response

  def __call__(self, **kwargs) -> lf.Message:
    lm_response = self.lm_response
    if lm_response is None:
      lm_response = super().__call__(**kwargs)
      if self.cache_response:
        self.rebind(lm_response=lm_response, skip_notification=True)
    return lm_response

  def clear_lm_response(self):
    """Clear LM response."""
    self.rebind(lm_response=None, skip_notification=True)



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common metrics for Langfun evaluation."""


import abc
import collections
import threading
from typing import Annotated, Any

from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import metric_values
import pyglove as pg


Rate = metric_values.Rate
Average = metric_values.Average


class Metric(pg.Object, pg.views.HtmlTreeView.Extension):
  """Interface for an evaluation metric."""

  name: Annotated[
      str,
      (
          'Name of the metric, which will be used as the key in the dict '
          'returned by `Experiment.metric_values()`'
      )
  ]

  def _on_bound(self):
    super()._on_bound()
    self._label_group = None
    self._lock = threading.Lock()

  def audit(self, example: example_lib.Example) -> dict[str, Any]:
    """Audits a processed example and returns metric metadata for it."""
    # NOTE(daiyip): the metric values are being updated concurrently, so we
    # uses a lock to avoid race condition. We might consider relaxing the lock
    # later if metric auditing becomes a bottleneck.
    with self._lock:
      for v in self.values():
        v.increment_total()

      metadata = self._audit(example)

      self._update_view()
      return metadata

  @abc.abstractmethod
  def _audit(self, example: example_lib.Example) -> dict[str, Any]:
    """Subclasses should override this method to implement the metric logic."""

  @abc.abstractmethod
  def values(self) -> list[metric_values.MetricValue]:
    """Returns all the values computed by this metric."""

  def reset(self) -> None:
    """Resets the metric values."""
    for v in self.values():
      v.reset()

  def _update_view(self):
    """Refreshes the metric values."""
    if self._label_group is None:
      return

    for label, value in zip(self._label_group.labels, self.values()):
      label.update(
          text=self._metric_value_text(value),
          tooltip=self._metric_value_tooltip(value),
      )

  def _metric_value_text(self, metric_value: metric_values.MetricValue) -> str:
    """Returns the label text for the metric value."""
    return str(metric_value)

  def _metric_value_tooltip(
      self, metric_value: metric_values.MetricValue) -> str:
    """Returns the label text for the metric value."""
    with pg.str_format(verbose=True):
      return f'{metric_value.sym_path.key}: {metric_value}'

  def _metric_label_text(self) -> str:
    return ''.join(
        c for c in self.__class__.__name__
        if c.isalnum() and not c.islower()
    )

  def _metric_label_tooltip(self) -> str:
    return self.__class__.__type_name__

  def _html_tree_view(
      self,
      *,
      view: pg.views.HtmlTreeView,
      extra_flags: dict[str, Any] | None = None,
      **kwargs,
  ) -> pg.Html:
    """Renders the content of the metric value."""
    extra_flags = extra_flags or {}
    interactive = extra_flags.get('interactive', True)
    label_group = self._label_group
    if label_group is None:
      label_group = pg.views.html.controls.LabelGroup(
          [
              pg.views.html.controls.Label(
                  self._metric_value_text(mv),
                  tooltip=self._metric_value_tooltip(mv),
                  css_classes=[mv.sym_path.key, 'metric-value'],
                  interactive=interactive,
              ) for mv in self.values()
          ],
          name=pg.views.html.controls.Label(
              self._metric_label_text(),
              tooltip=self._metric_label_tooltip(),
              css_classes=[
                  'metric-name',
                  pg.object_utils.camel_to_snake(self.__class__.__name__, '-')
              ],
              interactive=False,
          ),
          css_classes=['metric-container'],
      )
      if interactive:
        self._label_group = label_group
    return label_group.to_html()

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .metric-container {
          display: inline-flex;
          overflow: hidden;
          border-radius: 5px;
          border: 0px;
          margin: 5px;
          padding: 0px;
        }
        .metric-container .label-container {
          vertical-align: middle;
        }
        .metric-value.oop_errors {
          color: magenta;
          background-color: #f9e6eb;
        }
        .metric-value.non_oop_errors {
          color: red;
          background-color: #fdcccc;
        }
        """
    ]

#
# Common metrics.
#


class MetricBase(Metric):
  """Base class for common metrics."""

  oop_errors: Rate | None = Rate()
  non_oop_errors: Rate | None = Rate()

  def _on_bound(self) -> None:
    super()._on_bound()
    self._error_breakdown = collections.defaultdict(list)

  def reset(self) -> None:
    """Resets the metric."""
    super().reset()
    self._error_breakdown = collections.defaultdict(list)

  def _audit(self, example: example_lib.Example) -> dict[str, Any]:
    """Audits the evaluation example after processing."""
    if example.error is None:
      return self._audit_processed(example)
    else:
      return self._audit_error(example)

  def _audit_error(self, example: example_lib.Example) -> dict[str, Any]:
    """Audits the evaluation example after processing."""
    assert example.error is not None
    tag = example.error.tag
    if tag.startswith('MappingError'):
      self.oop_errors.add(example.id, 1)
    else:
      self.non_oop_errors.add(example.id, 1)
    self._error_breakdown[tag].append(example.id)
    return dict(error=tag)

  @abc.abstractmethod
  def _audit_processed(self, example: example_lib.Example) -> dict[str, Any]:
    """Audits the evaluation example after processing."""

  def _oop_errors_breakdown(self) -> str | None:
    """Returns the OOP error breakdown as a string."""
    return '\n'.join(
        [
            f'- {k}: {len(v)}' for k, v in self._error_breakdown.items()
            if k.startswith('MappingError')
        ]
    ) or None

  def _non_oop_errors_breakdown(self) -> str | None:
    """Returns the non-OOP error breakdown as a string."""
    return '\n'.join(
        [
            f'- {k}: {len(v)}' for k, v in self._error_breakdown.items()
            if not k.startswith('MappingError')
        ]
    ) or None

  def _sym_nondefault(self) -> dict[str, Any]:
    """Overrides nondefault valuesso volatile values are not included."""
    return dict()


class Match(MetricBase):
  """Metric for matching outputs against groundtruth."""

  name = 'match'
  matches: Rate = Rate()
  mismatches: Rate = Rate()

  def match(
      self, example_input: Any, output: Any
  ) -> bool | tuple[bool, dict[str, Any]]:
    """Returns whether the output matches the groundtruth from the example.

    Args:
      example_input: The example input which contains the groundtruth.
      output: The output to match against.

    Returns:
      True if the output matches the groundtruth, False otherwise.
      Or a tuple of (match, metadata).
    """
    groundtruth = getattr(example_input, 'groundtruth', pg.MISSING_VALUE)
    if pg.MISSING_VALUE == groundtruth:
      raise ValueError(
          f'`groundtruth` is not present in the example ({example_input}). '
          'Please subclassing `Match` and override the `match` method to '
          'support custom example format.'
      )
    return pg.eq(output, groundtruth)

  def _audit_processed(self, example: example_lib.Example) -> dict[str, Any]:
    """Audits the evaluation example after processing."""
    metadata = {}
    is_match = self.match(example.input, example.output)
    if isinstance(is_match, tuple):
      is_match, metadata = is_match
    if is_match:
      self.matches.add(example.id, 1)
      metadata['match'] = True
    else:
      self.mismatches.add(example.id, 1)
      metadata['mismatch'] = True
    return metadata

  def values(self) -> list[metric_values.MetricValue]:
    """Returns all the values computed by this metric."""
    return [
        self.matches,
        self.mismatches,
        self.oop_errors,
        self.non_oop_errors
    ]

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .metric-name.match {
          padding: 5px;
          color: white;
          background-color: purple;
        }
        .metric-value.matches {
          color: green;
          background-color: #dcefbe;
        }
        .metric-value.mismatches {
          color: orange;
          background-color: #ffefc4;
        }
        """
    ]


class Score(MetricBase):
  """Base class for scoring."""

  name = 'score'
  average_score: Average = Average()

  @abc.abstractmethod
  def score(
      self,
      example_input: Any,
      output: Any) -> float | tuple[float, dict[str, Any]]:
    """Returns the score based on the example and output.

    Args:
      example_input: The example input based on which the output is generated.
      output: The output to score.

    Returns:
      A float score. Or a tuple of (score, metadata).
    """

  def _audit_processed(self, example: example_lib.Example) -> dict[str, Any]:
    """Audits the evaluation example after processing."""
    metadata = {}
    score = self.score(example.input, example.output)
    if isinstance(score, tuple):
      score, metadata = score
    self.average_score.add(example.id, score)
    metadata['score'] = score
    return metadata

  def values(self) -> list[metric_values.MetricValue]:
    """Returns all the values computed by this metric."""
    return [
        self.average_score,
        self.oop_errors,
        self.non_oop_errors
    ]

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        .metric-name.score {
          padding: 5px;
          color: white;
          background-color: blue;
        }
        .metric-value.average_score {
          color: blue;
          background-color: #b0c7f6;
        }
        """
    ]



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Reporting evaluation results."""

import os
import threading
import time
import traceback
from typing import Annotated

from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
import pyglove as pg

Runner = experiment_lib.Runner
Experiment = experiment_lib.Experiment
Example = example_lib.Example


_SUMMARY_FILE = 'summary.html'
_EVALULATION_DETAIL_FILE = 'index.html'


class HtmlReporter(experiment_lib.Plugin):
  """Plugin for periodically generating HTML reports for the experiment."""

  summary_interval: Annotated[
      int,
      'The interval of writing summary in seconds.'
  ] = 60

  experiment_report_interval: Annotated[
      int,
      'The interval of writing report for inidividual experiments in seconds.'
  ] = 120

  def _on_bound(self):
    super()._on_bound()
    self._last_summary_time = 0
    self._last_experiment_report_time = {}
    self._update_thread = None
    self._stop_update = False
    self._stop_update_experiment_ids = set()
    self._summary_lock = None
    self._experiment_index_lock = None

  def on_run_start(
      self,
      runner: Runner,
      root: Experiment
  ) -> None:
    self._last_experiment_report_time = {leaf.id: 0 for leaf in root.leaf_nodes}
    self._stop_update = False
    self._stop_update_experiment_ids = set()
    self._summary_lock = threading.Lock()
    self._experiment_index_lock = {
        leaf.id: threading.Lock() for leaf in root.leaf_nodes
    }
    self._update_thread = threading.Thread(
        target=self._update_thread_func, args=(runner,)
    )
    self._update_thread.start()

  def on_run_complete(
      self,
      runner: Runner,
      root: Experiment
  ) -> None:
    self._stop_update = True
    self._maybe_update_summary(runner, force=True)

  def on_run_abort(
      self,
      runner: Runner,
      root: Experiment,
      error: BaseException
  ) -> None:
    self._stop_update = True
    self._maybe_update_summary(runner, force=True)

  def _update_thread_func(self, runner: Runner):
    while not self._stop_update:
      self._maybe_update_summary(runner, background=False)
      for leaf in runner.current_run.experiment.leaf_nodes:
        if leaf.id in self._stop_update_experiment_ids:
          continue
        self._maybe_update_experiment_html(runner, leaf, background=False)
        if leaf.progress.is_stopped:
          self._stop_update_experiment_ids.add(leaf.id)
      time.sleep(5)

  def on_experiment_start(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    if experiment.is_leaf:
      self._maybe_update_experiment_html(runner, experiment)

  def on_experiment_complete(
      self, runner: Runner, experiment: Experiment
  ):
    if experiment.is_leaf:
      self._maybe_update_experiment_html(runner, experiment, force=True)

  def on_experiment_abort(
      self,
      runner: Runner,
      experiment: Experiment,
      error: BaseException
  ) -> None:
    del error
    assert experiment.is_leaf
    self._maybe_update_experiment_html(runner, experiment, force=True)

  def on_example_complete(
      self, runner: Runner, experiment: Experiment, example: Example
  ):
    self._save_example_html(runner, experiment, example)
    self._maybe_update_experiment_html(runner, experiment)
    self._maybe_update_summary(runner)

  def _maybe_update_summary(
      self,
      runner: Runner,
      background: bool = True,
      force: bool = False) -> None:
    """Maybe update the summary of current run."""
    run = runner.current_run
    def _summary():
      html = run.experiment.to_html(
          collapse_level=None,
          extra_flags=dict(
              current_run=run, interactive=False, card_view=True,
          )
      )
      with self._summary_lock:
        html.save(os.path.join(run.output_root, _SUMMARY_FILE))

    if force or (time.time() - self._last_summary_time > self.summary_interval):
      self._last_summary_time = time.time()
      if background:
        runner.background_run(_summary)
      else:
        _summary()

  def _maybe_update_experiment_html(
      self,
      runner: Runner,
      experiment: Experiment,
      force: bool = False,
      background: bool = True,
  ) -> None:
    def _save():
      index_html_path = runner.current_run.output_path_for(
          experiment, _EVALULATION_DETAIL_FILE
      )
      try:
        with pg.timeit() as t:
          html = experiment.to_html(
              collapse_level=None,
              extra_flags=dict(
                  current_run=runner.current_run,
                  interactive=False,
                  card_view=False,
              ),
          )
          with self._experiment_index_lock[experiment.id]:
            html.save(index_html_path)
          experiment.info(
              f'Updated {index_html_path!r} in {t.elapse:.2f} seconds.',
          )
      except BaseException as e:  # pylint: disable=broad-except
        experiment.error(
            f'Failed to generate {index_html_path!r}. '
            f'Error: {e}, Stacktrace: \n{traceback.format_exc()}.',
        )
        raise e

    if force or (
        time.time() - self._last_experiment_report_time[experiment.id]
        > self.experiment_report_interval
    ):
      self._last_experiment_report_time[experiment.id] = time.time()
      if background:
        runner.background_run(_save)
      else:
        _save()

  def _save_example_html(
      self, runner: Runner, experiment: Experiment, example: Example
  ) -> None:
    """Saves the example in HTML format."""
    current_run = runner.current_run
    def _generate():
      try:
        with pg.timeit() as t:
          html = example.to_html(
              collapse_level=None,
              enable_summary_tooltip=False,
              extra_flags=dict(
                  # For properly rendering the next link.
                  num_examples=getattr(experiment, 'num_examples', None)
              ),
          )
          html.save(
              runner.current_run.output_path_for(
                  experiment, f'{example.id}.html'
              )
          )
        experiment.info(
            f'\'{example.id}.html\' generated in {t.elapse:.2f} seconds. '
        )
      except BaseException as e:  # pylint: disable=broad-except
        experiment.error(
            f'Failed to generate \'{example.id}.html\'. '
            f'Error: {e}, Stacktrace: \n{traceback.format_exc()}.',
        )
        raise e

    def _copy():
      src_file = current_run.input_path_for(experiment, f'{example.id}.html')
      dest_file = current_run.output_path_for(experiment, f'{example.id}.html')

      if src_file == dest_file:
        return

      if not pg.io.path_exists(src_file):
        experiment.warning(
            f'Skip copying \'{example.id}.html\' as '
            f'{src_file!r} does not exist.'
        )
        return

      try:
        with pg.timeit() as t, pg.io.open(src_file, 'r') as src:
          content = src.read()
          with pg.io.open(dest_file, 'w') as dest:
            dest.write(content)
        experiment.info(
            f'\'{example.id}.html\' copied in {t.elapse:.2f} seconds.'
        )
      except BaseException as e:  # pylint: disable=broad-except
        experiment.error(
            f'Failed to copy {src_file!r} to {dest_file!r}. Error: {e}.'
        )
        raise e

    generate_example_html = current_run.generate_example_html
    if (generate_example_html == 'all'
        or (generate_example_html == 'new' and example.newly_processed)
        or (isinstance(generate_example_html, list)
            and example.id in generate_example_html)):
      op = _generate
    else:
      op = _copy
    runner.background_run(op)



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base memory test."""

import unittest
import langfun.core as lf
from langfun.core.memories.conversation_history import ConversationHistory
import pyglove as pg


class ConversationHistoryTest(unittest.TestCase):

  def test_history_with_no_round_limit(self):
    m = ConversationHistory()
    m.remember(('hi', 'hello'))
    m.remember(
        ('how are you', 'Fine, thank you. Anything I can help with?'),
    )
    m.remember(
        ('Not for now, bye.', 'Okay, bye!'),
    )
    self.assertEqual(len(m.turns), 3)
    self.assertEqual(len(m.messages), 6)
    self.assertTrue(
        pg.eq(
            m.recollect(),
            lf.Template(
                """
                User: hi
                AI: hello
                User: how are you
                AI: Fine, thank you. Anything I can help with?
                User: Not for now, bye.
                AI: Okay, bye!
                """
            )
        )
    )

  def test_history_with_round_limit(self):
    m = ConversationHistory(max_turns=1)
    m.remember(('hi', 'hello'))
    m.remember(
        ('how are you', 'Fine, thank you. Anything I can help with?'),
    )
    m.remember(
        ('Not for now, bye.', 'Okay, bye!'),
    )
    self.assertEqual(len(m.turns), 1)
    self.assertEqual(len(m.messages), 2)
    self.assertTrue(
        pg.eq(
            m.recollect(),
            lf.Template(
                """
                User: Not for now, bye.
                AI: Okay, bye!
                """
            )
        )
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Checkpointing evaluation runs."""
import abc
import re
import threading
import traceback
from typing import Annotated

import langfun.core as lf
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib
import pyglove as pg

Example = example_lib.Example
Experiment = experiment_lib.Experiment
Runner = experiment_lib.Runner


class Checkpointer(experiment_lib.Plugin):
  """Base class for checkpointing evaluation examples."""

  checkpoint_filename: Annotated[
      str,
      'Checkpoint file pattern.'
  ] = 'checkpoint.bagz'

  def on_experiment_start(
      self,
      runner: Runner,
      experiment: Experiment
  ) -> None:
    if not experiment.is_leaf:
      return

    current_run = runner.current_run
    if current_run.reprocess is not True:  # pylint: disable=g-bool-id-comparison
      if current_run.input_root != current_run.output_root:
        experiment.info(
            f'Warm starting from directory: {current_run.input_root}.'
        )
      self._load_experiment(runner, experiment)

    example_ids_to_evaluate = current_run.examples_to_evaluate(experiment)
    if experiment.state.ckpt_examples:
      loaded_example_ids = list(
          sorted(experiment.state.ckpt_examples.keys())
      )
      example_ids_to_evaluate -= set(loaded_example_ids)
      example_ids_to_evaluate = list(sorted(example_ids_to_evaluate))
      experiment.info(
          f'{len(experiment.state.ckpt_examples)} examples '
          'loaded from checkpoint files. Their outputs will be used '
          f'for recomputing metrics. Example IDs: {loaded_example_ids}.'
      )
      experiment.info(
          f'{len(example_ids_to_evaluate)} examples will be processed from '
          f'scratch. Example IDs: {example_ids_to_evaluate}.'
      )
    else:
      experiment.info(
          'No examples are loaded from checkpoint files. '
          f'{len(example_ids_to_evaluate)} examples will be processed from '
          f'scratch. Example IDs: {example_ids_to_evaluate}.'
      )

  def on_example_complete(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves the example to the checkpoint file."""
    if example.newly_processed:
      self._save_example(runner, experiment, example)

  def _load_experiment(
      self,
      runner: Runner,
      experiment: Experiment,
  ) -> None:
    """Creates the checkpoint file."""
    ckpt_files = self._list_checkpoint_filenames(runner, experiment)
    experiment.info(f'Found {len(ckpt_files)} checkpoint files to load.')

    # Load the checkpoint files in parallel.
    current_run = runner.current_run
    examples_to_load = current_run.examples_to_load(experiment)
    examples_to_load_metadata = current_run.examples_to_load_metadata(
        experiment
    )
    context = dict(counter=0, counter_lock=threading.Lock())
    copy_ckpt = current_run.input_root != current_run.output_root

    def _load_state(ckpt_file):
      error = None
      with pg.timeit() as t:
        try:
          experiment.load_state(
              current_run.input_path_for(experiment, ckpt_file),
              filter=lambda x: x.id in examples_to_load,
              load_example_metadata=lambda x: x.id in examples_to_load_metadata,
          )
        except BaseException as e:  # pylint: disable=broad-except
          error = e
        finally:
          with context['counter_lock']:
            context['counter'] += 1

          progress_str = f'{context["counter"]}/{len(ckpt_files)}'
          if error is None:
            experiment.info(
                f'Checkpoint file {ckpt_file!r} loaded in {t.elapse:.2f} '
                f'seconds. ({progress_str})'
            )
          else:
            experiment.warning(
                f'Failed to load checkpoint file {ckpt_file!r}: {error}. '
                f'Skipping the file. ({progress_str})'
            )

        if not copy_ckpt:
          return

        # Copy the checkpoint records to the output directory.
        try:
          with pg.io.open_sequence(
              current_run.output_path_for(experiment, ckpt_file), 'w'
          ) as o, pg.io.open_sequence(
              current_run.input_path_for(experiment, ckpt_file), 'r'
          ) as i:
            for x in i:
              o.add(x)
        except BaseException as e:  # pylint: disable=broad-except
          experiment.warning(
              f'Failed to copy checkpoint {ckpt_file!r}: {e}.'
          )

    _ = list(
        lf.concurrent_map(
            _load_state, ckpt_files, max_workers=16, silence_on_errors=None
        )
    )

  @abc.abstractmethod
  def _list_checkpoint_filenames(
      self, runner: Runner, experiment: Experiment
  ) -> list[str]:
    """Lists the checkpoint filenames to restore."""

  @abc.abstractmethod
  def _save_example(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves an evaluated example."""


class PerExampleCheckpointer(Checkpointer):
  """Checkpointer that saves each example to a separate file."""

  def _on_bound(self):
    super()._on_bound()
    prefix, ext = self._file_prefix_and_ext(self.checkpoint_filename)
    self._checkpoint_file_prefix = prefix
    self._checkpoint_file_ext = ext

  def _list_checkpoint_filenames(
      self, runner: Runner, experiment: Experiment
  ) -> list[str]:
    experiment_dir = runner.current_run.input_dir(experiment)
    filenames = []
    examples_to_load = runner.current_run.examples_to_load(experiment)
    if pg.io.path_exists(experiment_dir):
      regex = re.compile(
          f'{self._checkpoint_file_prefix}_(\\d+){self._checkpoint_file_ext}'
          .replace('.', '\\.')
      )
      for filename in pg.io.listdir(experiment_dir):
        match = regex.match(filename)
        if match and int(match.group(1)) in examples_to_load:
          filenames.append(filename)
    return filenames

  def _save_example(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves the example to the checkpoint file."""
    def save_state(example: Example):
      writer = SequenceWriter(
          runner.current_run.output_path_for(
              experiment,
              (
                  f'{self._checkpoint_file_prefix}_{example.id}'
                  f'{self._checkpoint_file_ext}'
              )
          )
      )
      try:
        writer.add(example)
        writer.close()
        experiment.info(
            f'Example {example.id} checkpointed to {writer.path}.',
        )
      except BaseException as e:  # pylint: disable=broad-except
        experiment.error(
            f'Failed to checkpoint example {example.id} to {writer.path}. '
            f'Error: {e}, Stacktrace: \n{traceback.format_exc()}.',
        )
        raise e
    runner.background_run(save_state, example)

  def _file_prefix_and_ext(self, filename: str) -> tuple[str, str]:
    ext_index = filename.rfind('.')
    if ext_index == -1:
      return filename, ''
    else:
      return filename[:ext_index], filename[ext_index:]


class BulkCheckpointer(Checkpointer):
  """Checkpointer that saves all examples to a single file."""

  def _on_bound(self):
    super()._on_bound()
    self._lock = threading.Lock()
    self._sequence_writer = None

  def on_run_start(
      self,
      runner: Runner,
      root: Experiment,
  ) -> None:
    self._sequence_writer = {}

  def on_run_abort(
      self,
      runner: Runner,
      root: Experiment,
      error: BaseException
  ) -> None:
    with self._lock:
      if self._sequence_writer is not None:
        for writer in self._sequence_writer.values():
          writer.close()
        self._sequence_writer.clear()

  def on_run_complete(
      self,
      runner: Runner,
      root: Experiment,
  ) -> None:
    with self._lock:
      assert self._sequence_writer is not None and not self._sequence_writer

  def on_experiment_start(
      self,
      runner: Runner,
      experiment: Experiment,
  ) -> None:
    super().on_experiment_start(runner, experiment)

    # Prepare the sequence writer for the experiment.
    if experiment.is_leaf:
      sequence_writer = SequenceWriter(
          runner.current_run.output_path_for(
              experiment, self.checkpoint_filename
          )
      )
      with self._lock:
        if self._sequence_writer is not None:
          self._sequence_writer[experiment.id] = sequence_writer

  def _list_checkpoint_filenames(
      self, runner: Runner, experiment: Experiment
  ) -> list[str]:
    if pg.io.path_exists(
        runner.current_run.input_path_for(experiment, self.checkpoint_filename)
    ):
      return [self.checkpoint_filename]
    return []

  def on_experiment_complete(
      self,
      runner: Runner,
      experiment: Experiment,
  ) -> None:
    """Closes the checkpoint file."""
    if not experiment.is_leaf:
      return
    assert experiment.id in self._sequence_writer
    with self._lock:
      if self._sequence_writer is not None:
        # Make sure the writer is closed without delay so the file will be
        # available immediately.
        writer = self._sequence_writer.pop(experiment.id)
        writer.close()
        experiment.info(
            f'{len(experiment.state.evaluation_status)} examples are '
            f'checkpointed to {writer.path}.'
        )

  def _save_example(
      self,
      runner: Runner,
      experiment: Experiment,
      example: Example,
  ) -> None:
    """Saves the example to the checkpoint file."""
    assert experiment.id in self._sequence_writer
    def _save_example(example: Example):
      writer = self._sequence_writer[experiment.id]
      try:
        writer.add(example)
        experiment.info(
            f'Example {example.id} checkpointed to {writer.path}.',
        )
      except BaseException as e:  # pylint: disable=broad-except
        experiment.error(
            f'Failed to checkpoint example {example.id} to {writer.path}. '
            f'Error: {e}, Stacktrace: \n{traceback.format_exc()}.',
        )
        raise e
    runner.background_run(_save_example, example)


class SequenceWriter:
  """Thread safe sequence writer."""

  def __init__(self, path: str):
    self._lock = threading.Lock()
    self._path = path
    self._sequence_writer = pg.io.open_sequence(path, 'a')

  @property
  def path(self) -> str:
    return self._path

  def add(self, example: Example):
    example_blob = pg.to_json_str(
        example,
        hide_default_values=True,
        save_ref_value=True,
        exclude_input=False,
    )
    with self._lock:
      if self._sequence_writer is None:
        return
      self._sequence_writer.add(example_blob)

  def close(self):
    # Make sure there is no write in progress.
    with self._lock:
      if self._sequence_writer is None:
        return
      self._sequence_writer.close()
      self._sequence_writer = None

  def __del__(self):
    self.close()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test Fake LLMs."""

import contextlib
import io
import unittest
import langfun.core as lf
from langfun.core.llms import fake as fakelm


class PseudoTest(unittest.TestCase):

  def test_sample(self):
    lm = fakelm.Pseudo()
    self.assertEqual(lm.model_id, 'Pseudo')
    with self.assertRaises(ValueError):
      _ = lm.sample(['hi'])


class EchoTest(unittest.TestCase):

  def test_sample(self):
    lm = fakelm.Echo()
    self.assertEqual(
        lm.sample(['hi']),
        [
            lf.LMSamplingResult(
                [
                    lf.LMSample(
                        lf.AIMessage(
                            'hi',
                            score=1.0,
                            logprobs=None,
                            is_cached=False,
                            usage=lf.LMSamplingUsage(2, 2, 4),
                            tags=[lf.Message.TAG_LM_RESPONSE],
                        ),
                        score=1.0,
                        logprobs=None,
                    )
                ],
                lf.LMSamplingUsage(2, 2, 4))
        ]
    )

  def test_call(self):
    string_io = io.StringIO()
    lm = fakelm.Echo(debug=True)
    with contextlib.redirect_stdout(string_io):
      self.assertEqual(lm('hi'), 'hi')
    debug_info = string_io.getvalue()
    self.assertIn('[0] LM INFO', debug_info)
    self.assertIn('[0] PROMPT SENT TO LM', debug_info)
    self.assertIn('[0] LM RESPONSE', debug_info)

  def test_score(self):
    lm = fakelm.Echo()
    self.assertEqual(
        lm.score('hi', ['hello', 'how are you']),
        [lf.LMScoringResult(0.0), lf.LMScoringResult(-1.0)],
    )

  def test_tokenize(self):
    lm = fakelm.Echo()
    self.assertEqual(
        lm.tokenize('hi'),
        [('hi', 0)]
    )


class StaticResponseTest(unittest.TestCase):

  def test_sample(self):
    canned_response = "I'm sorry, I can't help you with that."
    lm = fakelm.StaticResponse(canned_response)
    self.assertEqual(
        lm.sample(['hi']),
        [
            lf.LMSamplingResult(
                [
                    lf.LMSample(
                        lf.AIMessage(
                            canned_response,
                            score=1.0,
                            logprobs=None,
                            is_cached=False,
                            usage=lf.LMSamplingUsage(2, 38, 40),
                            tags=[lf.Message.TAG_LM_RESPONSE],
                        ),
                        score=1.0,
                        logprobs=None,
                    )
                ],
                usage=lf.LMSamplingUsage(2, 38, 40)
            )
        ],
    )
    self.assertEqual(
        lm.sample(['Tell me a joke.']),
        [
            lf.LMSamplingResult(
                [
                    lf.LMSample(
                        lf.AIMessage(
                            canned_response,
                            score=1.0,
                            logprobs=None,
                            is_cached=False,
                            usage=lf.LMSamplingUsage(15, 38, 53),
                            tags=[lf.Message.TAG_LM_RESPONSE],
                        ),
                        score=1.0,
                        logprobs=None,
                    )
                ],
                usage=lf.LMSamplingUsage(15, 38, 53)
            )
        ],
    )

  def test_call(self):
    string_io = io.StringIO()
    canned_response = "I'm sorry, I can't help you with that."
    lm = fakelm.StaticResponse(canned_response, debug=True)

    with contextlib.redirect_stdout(string_io):
      self.assertEqual(lm('hi'), canned_response)

    debug_info = string_io.getvalue()
    self.assertIn('[0] LM INFO', debug_info)
    self.assertIn('[0] PROMPT SENT TO LM', debug_info)
    self.assertIn('[0] LM RESPONSE', debug_info)


class StaticMappingTest(unittest.TestCase):

  def test_sample(self):
    lm = fakelm.StaticMapping({
        'Hi': 'Hello',
        'How are you?': 'I am fine, how about you?',
    }, temperature=0.5)
    self.assertEqual(lm.sampling_options.temperature, 0.5)
    self.assertEqual(
        lm.sample(['Hi', 'How are you?']),
        [
            lf.LMSamplingResult(
                [
                    lf.LMSample(
                        lf.AIMessage(
                            'Hello',
                            score=1.0,
                            logprobs=None,
                            is_cached=False,
                            usage=lf.LMSamplingUsage(2, 5, 7),
                            tags=[lf.Message.TAG_LM_RESPONSE],
                        ),
                        score=1.0,
                        logprobs=None,
                    )
                ],
                usage=lf.LMSamplingUsage(2, 5, 7)
            ),
            lf.LMSamplingResult(
                [
                    lf.LMSample(
                        lf.AIMessage(
                            'I am fine, how about you?',
                            score=1.0,
                            logprobs=None,
                            is_cached=False,
                            usage=lf.LMSamplingUsage(12, 25, 37),
                            tags=[lf.Message.TAG_LM_RESPONSE],
                        ),
                        score=1.0,
                        logprobs=None,
                    )
                ],
                usage=lf.LMSamplingUsage(12, 25, 37)
            )
        ]
    )
    with self.assertRaises(KeyError):
      _ = lm.sample(['I am not in the table.'])


class StaticSequenceTest(unittest.TestCase):

  def test_sample(self):
    lm = fakelm.StaticSequence([
        'Hello',
        'I am fine, how about you?',
    ], temperature=0.5)
    self.assertEqual(lm.sampling_options.temperature, 0.5)
    self.assertEqual(
        lm.sample(['Hi', 'How are you?']),
        [
            lf.LMSamplingResult(
                [
                    lf.LMSample(
                        lf.AIMessage(
                            'Hello',
                            score=1.0,
                            logprobs=None,
                            is_cached=False,
                            usage=lf.LMSamplingUsage(2, 5, 7),
                            tags=[lf.Message.TAG_LM_RESPONSE],
                        ),
                        score=1.0,
                        logprobs=None,
                    )
                ],
                usage=lf.LMSamplingUsage(2, 5, 7)
            ),
            lf.LMSamplingResult(
                [
                    lf.LMSample(
                        lf.AIMessage(
                            'I am fine, how about you?',
                            score=1.0,
                            logprobs=None,
                            is_cached=False,
                            usage=lf.LMSamplingUsage(12, 25, 37),
                            tags=[lf.Message.TAG_LM_RESPONSE],
                        ),
                        score=1.0,
                        logprobs=None,
                    )
                ],
                usage=lf.LMSamplingUsage(12, 25, 37)
            )
        ]
    )
    with self.assertRaises(IndexError):
      _ = lm.sample(['No next one.'])


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import tempfile
import unittest

from langfun.core.eval.v2 import eval_test_helper
from langfun.core.eval.v2 import evaluation as evaluation_lib
from langfun.core.eval.v2 import example as example_lib
from langfun.core.eval.v2 import experiment as experiment_lib

import pyglove as pg

Example = example_lib.Example
Evaluation = evaluation_lib.Evaluation
RunId = experiment_lib.RunId
Run = experiment_lib.Run


class EvaluationTest(unittest.TestCase):

  def test_hyper_evaluation(self):
    exp = eval_test_helper.TestEvaluation(
        lm=eval_test_helper.TestLLM(offset=pg.oneof(range(3)))
    )
    self.assertFalse(exp.is_leaf)
    self.assertTrue(
        pg.eq(
            exp.children,
            [
                eval_test_helper.TestEvaluation(
                    lm=eval_test_helper.TestLLM(offset=0)
                ),
                eval_test_helper.TestEvaluation(
                    lm=eval_test_helper.TestLLM(offset=1)
                ),
                eval_test_helper.TestEvaluation(
                    lm=eval_test_helper.TestLLM(offset=2)
                ),
            ]
        )
    )
    self.assertEqual(exp.children[0].num_examples, 10)
    self.assertEqual(
        [c.is_leaf for c in exp.children],
        [True] * len(exp.children)
    )
    self.assertEqual(
        [r.resource_ids() for r in exp.leaf_nodes],
        [set(['test_llm:0']), set(['test_llm:1']), set(['test_llm:2'])]
    )

  def test_input(self):
    exp = eval_test_helper.TestEvaluation()
    self.assertEqual(exp.num_examples, 10)
    exp = eval_test_helper.TestEvaluation(
        inputs=eval_test_helper.test_inputs(None)
    )
    self.assertEqual(exp.num_examples, 20)
    @pg.functor
    def my_inputs():
      yield pg.Dict(x=1, y=2)
      yield pg.Dict(x=3, y=4)
    exp = eval_test_helper.TestEvaluation(inputs=my_inputs())
    self.assertEqual(exp.num_examples, 2)

  def test_evaluate(self):
    exp = eval_test_helper.TestEvaluation()
    example = exp.evaluate(Example(id=3))
    self.assertTrue(exp.state.get_status(3).evaluated)
    self.assertFalse(exp.state.get_status(3).in_progress)
    self.assertTrue(exp.state.get_status(3).newly_processed)
    self.assertFalse(exp.state.get_status(3).has_error)
    self.assertEqual(exp.state.in_progress_examples, {})
    self.assertTrue(example.newly_processed)
    self.assertEqual(example.input, pg.Dict(x=2, y=4, groundtruth=6))
    self.assertEqual(example.output, 6)
    self.assertIsNone(example.error)
    self.assertEqual(example.metadata, {})
    self.assertEqual(example.metric_metadata, dict(match=True))
    self.assertIsNotNone(example.usage_summary)
    self.assertGreater(example.usage_summary.total.total_tokens, 0)
    self.assertEqual(example.usage_summary.total.num_requests, 1)
    self.assertIsNotNone(example.execution_status)
    self.assertIsNotNone(example.start_time)
    self.assertIsNotNone(example.end_time)

    exp = eval_test_helper.TestEvaluation(lm=eval_test_helper.TestLLM(offset=1))
    example = exp.evaluate(3)
    self.assertTrue(example.newly_processed)
    self.assertEqual(example.input, pg.Dict(x=2, y=4, groundtruth=6))
    self.assertEqual(example.output, 7)
    self.assertIsNone(example.error)
    self.assertEqual(example.metadata, {})
    self.assertEqual(example.metric_metadata, dict(mismatch=True))

    with self.assertRaisesRegex(ValueError, 'x should not be 5'):
      _ = exp.evaluate(6, raise_if_has_error=True)
    example = exp.evaluate(6)
    self.assertTrue(example.newly_processed)
    self.assertEqual(example.input, pg.Dict(x=5, y=25, groundtruth=30))
    self.assertEqual(pg.MISSING_VALUE, example.output)
    self.assertEqual(example.error.tag, 'ValueError')
    self.assertEqual(example.metadata, {})
    self.assertEqual(example.metric_metadata, dict(error='ValueError'))

  def test_evaluate_withstate(self):
    eval_dir = os.path.join(tempfile.mkdtemp(), 'test_eval')
    pg.io.mkdirs(eval_dir, exist_ok=True)
    state_file = os.path.join(eval_dir, 'state.jsonl')
    with pg.io.open_sequence(state_file, 'w') as f:
      exp = eval_test_helper.TestEvaluation()
      example = exp.evaluate(3)
      self.assertTrue(example.newly_processed)
      self.assertEqual(example.input, pg.Dict(x=2, y=4, groundtruth=6))
      self.assertEqual(example.output, 6)
      self.assertEqual(len(exp.state.evaluation_status), 1)
      f.add(pg.to_json_str(example))

      example2 = exp.evaluate(6)
      self.assertTrue(example2.newly_processed)
      self.assertEqual(example2.input, pg.Dict(x=5, y=25, groundtruth=30))
      self.assertTrue(example2.has_error)
      self.assertEqual(len(exp.state.evaluation_status), 2)
      f.add(pg.to_json_str(example2))

    exp.reset()
    self.assertEqual(len(exp.state.ckpt_examples), 0)
    exp.load_state(state_file)
    self.assertEqual(len(exp.state.ckpt_examples), 2)
    self.assertEqual(len(exp.state.evaluation_status), 0)
    example = exp.evaluate(3)
    self.assertFalse(example.newly_processed)
    self.assertEqual(example.input, pg.Dict(x=2, y=4, groundtruth=6))
    self.assertEqual(example.output, 6)
    self.assertGreater(example.usage_summary.total.total_tokens, 0)
    self.assertGreater(example.usage_summary.cached.total.total_tokens, 0)
    self.assertEqual(example.usage_summary.cached.total.num_requests, 1)
    self.assertEqual(example.usage_summary.uncached.total.total_tokens, 0)
    self.assertEqual(example.usage_summary.uncached.total.num_requests, 0)
    self.assertIsNotNone(example.start_time)
    self.assertIsNotNone(example.end_time)
    self.assertIn('evaluate.process', example.execution_status)
    self.assertIn('evaluate.metric', example.execution_status)

    example2 = exp.evaluate(6, reevaluate_upon_previous_errors=False)
    self.assertFalse(example2.newly_processed)
    self.assertEqual(example2.input, pg.Dict(x=5, y=25, groundtruth=30))
    self.assertTrue(example2.has_error)
    self.assertIsNotNone(example2.start_time)
    self.assertIsNotNone(example2.end_time)
    self.assertIn('evaluate.process', example2.execution_status)
    self.assertIn('evaluate.metric', example2.execution_status)

    # Test load_state with filter.
    exp.reset()
    self.assertEqual(len(exp.state.ckpt_examples), 0)
    exp.load_state(state_file, filter=lambda x: x.id == 3)
    self.assertEqual(len(exp.state.ckpt_examples), 1)

    exp.reset()
    self.assertEqual(len(exp.state.ckpt_examples), 0)
    exp.load_state(state_file, filter=lambda x: x.id == 1)
    self.assertEqual(len(exp.state.ckpt_examples), 0)

  def test_html_view(self):
    exp = eval_test_helper.TestEvaluation()
    exp.debug('debug message')
    exp.info('info message')
    exp.warning('warning message', x=1)
    exp.error('error message', x=1)
    exp.fatal('fatal message')

    self.assertIn(
        exp.id,
        exp.to_html(extra_flags=dict(card_view=True, current_run=None)).content
    )
    self.assertIn(
        exp.id,
        exp.to_html(
            extra_flags=dict(
                card_view=False,
                current_run=Run(
                    root_dir='/tmp/test_run',
                    id=RunId.from_id('20241031_1'),
                    experiment=pg.Ref(exp),
                )
            )
        ).content
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The core langfun module.

Symbols with name `langfun.core.*` will be seen as `langfun.*` directly.
Please see //third_party/py/langfun/__init__.py for details.
"""

# pylint: disable=g-bad-import-order
# pylint: disable=g-importing-member
# pylint: disable=g-import-not-at-top

# Constants
from langfun.core.component import RAISE_IF_HAS_ERROR

# Interface for all langfun components.
from langfun.core.component import Component

from langfun.core.template import Template
from langfun.core.langfunc import LangFunc

# Decorator for set the positional init args for component.
from langfun.core.component import use_init_args

# Context manager for overriding the values for contextual attributes.
from langfun.core.component import context

# For backward compatibility.
as_context = context
use_context = context

# Invoke a callable object asynchronously.
from langfun.core.async_support import invoke_async

# Shortcut function for overriding components attributes, usually for
# override settings.
from langfun.core.component import use_settings

from langfun.core.component import get_contextual_override
from langfun.core.component import context_value

# Value marker for attribute whose values will be provided from parent
# objects or from the `pg.component_context` context manager.
from langfun.core.component import contextual

# Content subscriptions through LangFunc events and event handlers.
from langfun.core.subscription import Event
from langfun.core.subscription import EventHandler

from langfun.core.subscription import subscribe
from langfun.core.subscription import unsubscribe
from langfun.core.subscription import subscribers

from langfun.core.subscription import subscriptions
from langfun.core.subscription import clear_subscriptions

# Events
from langfun.core.template import TemplateRenderEvent
from langfun.core.langfunc import LangFuncCallEvent

# Helper methods for concurrent sampling.
from langfun.core.sampling import sweep
from langfun.core.sampling import random_sample

# Concurrent execute a function with parallel inputs with inheriting current
# context's defaults and overrides.
from langfun.core.concurrent import RetryEntry
from langfun.core.concurrent import concurrent_execute
from langfun.core.concurrent import concurrent_map
from langfun.core.concurrent import with_retry

# Interface for natural language formattable.
from langfun.core.natural_language import NaturalLanguageFormattable

# Input/output protocols.
from langfun.core.message import Message
from langfun.core.message import UserMessage
from langfun.core.message import AIMessage
from langfun.core.message import SystemMessage
from langfun.core.message import MemoryRecord

from langfun.core.message import MessageConverter

# Interface for modality.
from langfun.core.modality import Modality
from langfun.core.modality import ModalityRef
from langfun.core.modality import ModalityError

# Interfaces for languge models.
from langfun.core.language_model import ModelInfo
from langfun.core.language_model import LanguageModel

from langfun.core.language_model import LMSample
from langfun.core.language_model import LMSamplingOptions
from langfun.core.language_model import LMSamplingResult
from langfun.core.language_model import LMScoringResult

from langfun.core.language_model import LMSamplingUsage
from langfun.core.language_model import UsageNotAvailable
from langfun.core.language_model import UsageSummary

from langfun.core.language_model import LMCache
from langfun.core.language_model import LMDebugMode

from langfun.core.language_model import LMError
from langfun.core.language_model import LMInputError
from langfun.core.language_model import ContextLimitError
from langfun.core.language_model import RetryableLMError
from langfun.core.language_model import RateLimitError
from langfun.core.language_model import TemporaryLMError

# Context manager for tracking usages.
from langfun.core.language_model import track_usages

# Components for building agents.
from langfun.core.memory import Memory

# Utility for console output.
from langfun.core import console

# Utility for event logging.
from langfun.core import logging

# pylint: enable=g-import-not-at-top
# pylint: enable=g-importing-member
# pylint: enable=g-bad-import-order



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""langfun text template with compositionality."""

import contextlib
import dataclasses
import functools
import inspect
from typing import Annotated, Any, Callable, Iterator, Set, Tuple, Type, Union

import jinja2
from jinja2 import meta as jinja2_meta
from langfun.core import component
from langfun.core import message as message_lib
from langfun.core import modality
from langfun.core import natural_language
from langfun.core import subscription
import pyglove as pg


# Include this string anywhere in the class docstr in order not to treat
# the docstring as a template.
NO_TEMPLATE_DOCSTR_SIGN = 'THIS IS NOT A TEMPLATE'

# Keys for storing render stack in the
# thread-local storage.
_TLS_RENDER_STACK = '_template_render_stack'
_TLS_RENDER_RESULT_CACHE = '_template_render_result_cache'

# The prefix for fields or contextual attributes to be treated as additional
# metadata for rendered message.
_ADDITIONAL_METADATA_PREFIX = 'metadata_'


class Template(
    natural_language.NaturalLanguageFormattable,
    component.Component,
    pg.typing.CustomTyping,
    pg.views.HtmlTreeView.Extension
):
  """Langfun string template.

  Langfun uses jinja2 as its template engine. Pleaes check out
  https://jinja.palletsprojects.com/en/3.1.x/templates/ for detailed
  explanation on the template language.
  """

  template_str: Annotated[
      str,
      (
          'A template string in jinja2 syntax. During `render`, the variables '
          'will be resolved from 1) the `kwargs` dict passed to the `render` '
          'method; 2) the attributes of this object; and 3) the attributes of '
          'its containing objects.'
      ),
  ]

  clean: Annotated[
      bool,
      (
          'If True, `inspect.cleandoc` will be applied on `template_str` to '
          'additional indention, etc. Otherwise, the original form of the '
          'template string will be used.'
      ),
  ] = True

  __kwargs__: Annotated[
      Any,
      (
          'Wildcard keyword arguments for `__init__` that can be referred in '
          'the template string. This allows modularization of prompt with '
          'fully or partially bound variables.'
      ),
  ]

  def __init_subclass__(cls):
    # NOTE(daiyip): class attribute `template_str` may exists in 2 ways:
    # A string: specified by the user as the new default value for template_str.
    # A property: inherited from the base class. In this case, we only update
    #   its default value when template_str can be extracted from the docstr.
    # Here we try to detect `template_str` from docstr when it's not specified
    # by the user.
    template_str = getattr(cls, 'template_str', None)

    if not isinstance(template_str, str):
      # Here we try to update the default value of `template_str` based on
      # the class docstring. Unless it's not provided or
      # 'THIS IS NOT A TEMPLATE' appears in the docstring.
      template_str = cls._template_str_from_docstr()
      if template_str:
        # Set the `template_str` attribute will change the default value
        # of the `template_str` symbolic field.
        setattr(cls, 'template_str', template_str)

    if template_str:
      # Declare template variables as symbolic attributes.
      template_vars = Template.resolve_vars(template_str)
      for var_name in template_vars:
        if 'DEFAULT' == var_name:
          raise ValueError(
              '`{{ DEFAULT }}` cannot be used in pre-configured templates. '
              f'Encountered: {template_str!r}'
          )
        # NOTE(daiyip): This is to avoid warning from accessing
        # `pg.Object.schema`, which was replaced by `pg.Object.__schema__`.
        if var_name == 'schema' or not hasattr(cls, var_name):
          setattr(cls, var_name, component.contextual())

    super().__init_subclass__()

  @classmethod
  def _template_str_from_docstr(cls) -> str:
    """Extract template string from docstring."""
    docstr = cls.__doc__
    if (
        not docstr
        or docstr is cls.__base__.__doc__
        or NO_TEMPLATE_DOCSTR_SIGN in docstr
    ):
      return ''

    docstr = inspect.cleandoc(docstr)
    doc_start = docstr.find('\n')
    if doc_start == -1:
      return ''

    return docstr[doc_start + 1 :].strip()

  @classmethod
  def resolve_vars(cls, template_str: str) -> Set[str]:
    try:
      env = jinja2.Environment()
      ast = env.parse(template_str)
      return jinja2_meta.find_undeclared_variables(ast)
    except jinja2.TemplateSyntaxError as e:
      raise ValueError(f'Bad template string:\n\n{template_str}') from e

  def _on_bound(self) -> None:
    super()._on_bound()

    # Invalidate cached properties.
    if self.clean:
      self.rebind(
          template_str=inspect.cleandoc(self.template_str),
          skip_notification=True,
      )

    # Invalidate cached variables and template cache.
    self.__dict__.pop('_variables', None)
    self.__dict__.pop('_template', None)

    # Use contextual value for unassigned attributes.
    # TODO(daiyip): Consider to delay template parsing upon usage.
    unassigned_vars = {}
    for k in self._variables:
      if k not in ('DEFAULT',) and not hasattr(self, k):
        unassigned_vars[k] = component.contextual()
    if unassigned_vars:
      self.rebind(unassigned_vars, skip_notification=True)

    # Last render output.
    self._cached_render_output = None

  @property
  def render_output(self) -> message_lib.Message | None:
    """Returns last render output."""
    return self._cached_render_output

  @functools.cached_property
  def _variables(self) -> Set[str]:
    """Returns all declared variables from the template."""
    return Template.resolve_vars(self.template_str)

  @functools.cached_property
  def _template(self) -> jinja2.Template:
    return jinja2.Template(self.template_str)

  def vars(
      self,
      specified: bool | None = None,
      closure: bool = False,
      leaf: bool | None = None,
  ) -> Set[str]:
    """Returns referred variables.

    Args:
      specified: If True, include only variables that are specified. If False,
        include only variables that are not specified. If None, include both.
      closure: If True, include variables from referred LangFuncs recursively.
        Otherwise, include the immediate used variables.
      leaf: If True, include only the non-LangFunc variables. If False, include
        LangFunc variables. If None, include both.

    Returns:
      A list of variable names that match the criteria.
    """
    variables = set()
    for k in self._variables:
      v = getattr(self, k, pg.MISSING_VALUE)

      match specified:
        case None:
          include = True
        case True:
          include = v != pg.MISSING_VALUE
        case _:
          include = v == pg.MISSING_VALUE

      if include and leaf is not None:
        include = leaf != isinstance(v, Template)

      if include:
        variables.add(k)

      if closure and isinstance(v, Template):
        variables.update(v.vars(closure=True, specified=specified, leaf=leaf))
    return variables

  @property
  def missing_vars(self) -> Set[str]:
    """Returns the missing variable names."""
    return self.vars(closure=True, specified=False)

  @classmethod
  def raw_str(cls, text: str) -> str:
    """Returns a template string that preserve the text as original."""
    return '{% raw %}' + text + '{% endraw %}'

  @classmethod
  def from_raw_str(cls, text: str) -> 'Template':
    """Returns a template that preserve the text as original."""
    return cls(cls.raw_str(text), clean=False)

  def render(
      self,
      *,
      allow_partial: bool = False,
      implicit: bool = False,
      message_cls: Type[message_lib.Message] = message_lib.UserMessage,
      **kwargs,
  ) -> message_lib.Message:
    """Renders the template with variables from the context.

    Args:
      allow_partial: Allow partial rendering, this means that unresolved
        variables are allowed and remain in the output text.
      implicit: If True, reuse the rendering output if a parent LangFunc
        is rendering current LangFunc multiple times. This is important
        for making sure all references to the same LangFunc within a single
        top-level rendering would return the same result. If False, every call
        to `render` will trigger the actual rendering process.
      message_cls: The message class used for creating the return value.
      **kwargs: Values for template variables.

    Returns:
      An Message object as the rendered result.
    """
    try:
      pg.object_utils.thread_local_push(_TLS_RENDER_STACK, self)

      # NOTE(daiyip): Rendering could be triggered explicitly (by the user code)
      # or implicitly (when `str` gets called during template rendering). Since
      # a LangFunc object could be referenced in a template multiple times, we
      # need to have all the references to the same LangFunc object produce the
      # same output. Rendering each sub-LangFunc object once and only once would
      # also greatly improve the rendering performance.
      #
      # To achieve this, we install a result cache for every explicit rendering.
      # Thus, when sub-LangFunc are rendered, they could retrieve the output
      # from the cache.

      cache = {}
      if implicit:
        cache = pg.object_utils.thread_local_get(_TLS_RENDER_RESULT_CACHE, {})

        if id(self) in cache:
          return cache[id(self)]

        caching_context = contextlib.nullcontext()
      else:
        caching_context = pg.object_utils.thread_local_value_scope(
            _TLS_RENDER_RESULT_CACHE, cache, {}
        )

      with caching_context:
        # We use **kwargs for both the missing parts of the child components
        # and as the overriden attributes of current component.
        with component.context(**kwargs):
          with self.override(**kwargs):
            inputs = dict()
            for var_name in self._variables:
              var_value = getattr(self, var_name, pg.MISSING_VALUE)
              if var_value == pg.MISSING_VALUE:
                if allow_partial:
                  var_value = _UnresolvedExpression(var_name)
                else:
                  raise ValueError(
                      f'The value for template variable {var_name!r} is not '
                      f'provided. Template: {self.template_str!r}'
                  )
              inputs[var_name] = var_value

            # Enable Python format for builtin types during template rendering,
            # versus the default PyGlove format (e.g. [0: 'abc'] for list).
            # User-defined classes may have their own format.
            with pg.object_utils.str_format(
                # Use compact (single-line) Python format (vs. PyGlove format)
                # quoted with markdown notions (e.g. `Foo(1)`)
                # for non-natural-language-formattable symbolic objects.
                markdown=True,
                compact=True,
                python_format=True,
            ):
              # Natural language formattable objects will be returned in natural
              # language when they are directly returned as rendering elements
              # in the template.
              with modality.format_modality_as_ref():
                rendered_text = self._template.render(**inputs)

            # Carry additional metadata.
            metadata = self.additional_metadata()

        if self.clean:
          rendered_text = rendered_text.strip()

        metadata.update(
            {k: pg.Ref(v) for k, v in inputs.items() if not inspect.ismethod(v)}
        )

        # Fill the variables for rendering the template as metadata.
        message = message_cls(text=rendered_text, metadata=metadata)

        # Tag input as rendered message.
        message.tag(message_lib.Message.TAG_RENDERED)

        # Adding result to cache.
        cache[id(self)] = message

        # Set last render output to message.
        self._cached_render_output = message

        # Emit TemplateRenderEvent.
        # Event handlers could access the cached result from nested templates
        # using `str()`.
        render_stack = list(pg.object_utils.thread_local_get(_TLS_RENDER_STACK))
        render_stack.pop()
        subscription.emit(
            TemplateRenderEvent(
                sender=self, output=message, render_stack=render_stack
            )
        )
      return message
    finally:
      top = pg.object_utils.thread_local_pop(_TLS_RENDER_STACK)
      assert top is self, (top, self)

  def additional_metadata(self) -> dict[str, Any]:
    """Returns additional metadta to be carried in the rendered message."""
    metadata = {}
    # Carry metadata from `lf.context`.
    for k, v in component.all_contextual_values().items():
      if k.startswith(_ADDITIONAL_METADATA_PREFIX):
        metadata[k.removeprefix(_ADDITIONAL_METADATA_PREFIX)] = v

    # Carry metadata from fields.
    for k, v in self.sym_init_args.items():
      if k.startswith(_ADDITIONAL_METADATA_PREFIX):
        metadata[k.removeprefix(_ADDITIONAL_METADATA_PREFIX)] = v
    return metadata

  #
  # Implements `pg.typing.CustomTyping`.
  #

  def custom_apply(
      self,
      path: pg.KeyPath,
      value_spec: pg.typing.ValueSpec,
      allow_partial: bool,
      child_transform: Callable[[pg.KeyPath, pg.typing.Field, Any], Any]
      | None = None,
  ) -> Tuple[bool, Any]:
    """Makes it applicable to pg.typing.Str()."""
    del allow_partial
    del child_transform

    # Check if value_spec directly accepts `self`.
    if value_spec.value_type and isinstance(self, value_spec.value_type):
      return (False, self)

    pg.typing.ensure_value_spec(value_spec, pg.typing.Str().noneable(), path)
    return (False, self)

  #
  # Implements `lf.NaturalLanguageFormattable`.
  #

  def natural_language_format(self) -> str:
    """Returns the natural language format representation."""
    return self.render(allow_partial=True, implicit=True).text

  def __eq__(self, other: Any) -> bool:
    if isinstance(other, str):
      return not self.missing_vars and str(self) == other
    return super().__eq__(other)

  def __hash__(self) -> int:
    # Override __hash__ since __eq__ has changed.
    return object.__hash__(self)

  #
  # Special methods.
  #

  @property
  def DEFAULT(self) -> 'Template':
    """Referring to the default value used for this template.

    This method is intended to be used in template for referring to the default
    value of current template. There are two scenarios:

    Scenario 1: Use instance-level template_str to override the class default.

    ```
    class Foo(lf.Template):
       '''Foo template.

       This is {{x}}.
       '''

    f = Foo(template_str='<h1>{{DEFAULT}}</h1>', x=1)
    f.render()

    >> <h1>This is 1.</h1>
    ```

    Scenario 2: Use an ad-hoc template to override a predefined field.

    ```
    class Bar(lf.Template):
      '''Bar template.

      {{preamble}}
      {{prompt}}
      '''
      preamble: lf.Template = lf.Template('You are a chat bot.')
      prompt: lf.Template = lf.Template('User: hi')

    b = Bar(preamble=lf.Template('<h1>{{DEFAULT}}<h1>'),
            prompt=lf.Template('<h2>{{DEFAULT}}</h2>')
    b.render()

    >> <h1>You are a chat bot.<h1>
    >> <h2>User: hi</h2>
    ```

    Returns:
      The default (pre-configured) value used for this template.
    """
    base_template = self.__class__.__schema__['template_str'].default_value
    if base_template == pg.MISSING_VALUE:
      if not self.sym_path:
        raise ValueError(
            f'No DEFAULT template found for {self!r}: '
            'The template neither has a default `template_str` nor is '
            'contained under another object.'
        )
      key = self.sym_path.key
      assert self.sym_parent is not None
      assigned_field = self.sym_parent.sym_attr_field(key)
      container_cls = self.sym_parent.__class__

      if (
          assigned_field is None
          or assigned_field.default_value == pg.MISSING_VALUE
      ):
        raise ValueError(
            f'No DEFAULT template found for {self!r}: '
            f'`{container_cls.__name__}.{key}` '
            'does not have a default value. '
        )
      base_template = assigned_field.default_value
      if isinstance(base_template, Template):
        base_template = base_template.template_str
      if not isinstance(base_template, str):
        raise ValueError(
            f'No DEFAULT template found for {self!r}: The default '
            f'value {base_template!r} of '
            f'`{container_cls.__name__}.{key}` is not a '
            '`lf.Template` object or str.'
        )
    t = Template(base_template)
    # NOTE(daiyip): Set the parent of the newly created template to self so
    # it could access all the contextual variables.
    t.sym_setparent(self)
    return t

  @classmethod
  def from_value(
      cls,
      value: Union[str, message_lib.Message, 'Template'],
      **kwargs
  ) -> 'Template':
    """Create a template object from a string or template."""
    if isinstance(value, cls):
      return value.clone(override=kwargs) if kwargs else value  # pylint: disable=no-value-for-parameter
    if isinstance(value, str):
      return cls(template_str=value, **kwargs)
    if isinstance(value, message_lib.Message):
      kwargs.update(value.metadata)
      return cls(template_str=value.text, **kwargs)
    if isinstance(value, Template):
      lfun = cls(template_str=value.template_str, **kwargs)
      # So lfun could acccess all attributes from value.
      lfun.sym_setparent(value)
      return lfun
    return cls(template_str='{{input}}', input=value, **kwargs)

  def _html_tree_view_content(
      self,
      *,
      view: pg.views.HtmlTreeView,
      root_path: pg.KeyPath | None = None,
      collapse_level: int | None = None,
      extra_flags: dict[str, Any] | None = None,
      debug: bool = False,
      **kwargs,
  ):
    extra_flags = extra_flags if extra_flags is not None else {}
    collapse_template_vars_level: int | None = extra_flags.get(
        'collapse_template_vars_level', 1
    )

    def render_template_str():
      return pg.Html.element(
          'div',
          [
              pg.Html.element('span', [self.template_str])
          ],
          css_classes=['template-str'],
      )

    def render_fields():
      return view.complex_value(
          {k: v for k, v in self.sym_items()},
          name='fields',
          root_path=root_path,
          parent=self,
          exclude_keys=['template_str', 'clean'],
          collapse_level=max(
              collapse_template_vars_level, collapse_level
          ) if collapse_level is not None else None,
          extra_flags=extra_flags,
          debug=debug,
          **view.get_passthrough_kwargs(
              remove=['exclude_keys'],
              **kwargs,
          )
      )

    return pg.views.html.controls.TabControl([
        pg.views.html.controls.Tab(
            'template_str',
            render_template_str(),
        ),
        pg.views.html.controls.Tab(
            'variables',
            render_fields(),
        ),
    ], selected=1)

  @classmethod
  def _html_tree_view_css_styles(cls) -> list[str]:
    return super()._html_tree_view_css_styles() + [
        """
        /* Langfun Template styles. */
        .template-str {
            padding: 10px;
            margin: 10px 5px 10px 5px;
            font-style: italic;
            font-size: 1.1em;
            white-space: pre-wrap;
            border: 1px solid #EEE;
            border-radius: 5px;
            background-color: #EEE;
            color: #cc2986;
        }
        """
    ]

  @classmethod
  @functools.cache
  def _html_tree_view_config(cls) -> dict[str, Any]:
    return pg.views.HtmlTreeView.get_kwargs(
        super()._html_tree_view_config(),
        dict(
            css_classes=['lf-template'],
        )
    )


# Register converter from str to LangFunc, therefore we can always
# pass strs to attributes that accept LangFunc.
pg.typing.register_converter(str, Template, Template)


@dataclasses.dataclass
class TemplateRenderEvent(subscription.Event[Template]):
  """Template rendering event."""

  output: message_lib.Message
  render_stack: list[Template]


class _UnresolvedExpression(pg.Object):
  """Unresolved expression in a Jinja2 template for partial rendering."""

  expression: str

  def __repr__(self) -> str:
    return self.expression

  def __str__(self) -> str:
    return '{{' + self.expression + '}}'

  def __call__(self, *args, **kwargs) -> '_UnresolvedExpression':
    items = [repr(x) for x in args] + [f'{k}={v!r}' for k, v in kwargs.items()]
    arg_str = ', '.join(items)
    return _UnresolvedExpression(f'{self.expression}({arg_str})')

  def __getitem__(self, key: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression}[{key!r}]')

  def __getattr__(self, key: str) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression}.{key}')

  def __len__(self) -> '_UnresolvedExpression':     # pylint: disable=invalid-length-returned
    return _UnresolvedExpression(f'len({self.expression})')

  def __neg__(self) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'-{self.expression}')

  def __add__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} + {other!r}')

  def __radd__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{other!r} + {self.expression}')

  def __sub__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} - {other!r}')

  def __rsub__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{other!r} - {self.expression}')

  def __mul__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} * {other!r}')

  def __rmul__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{other!r} * {self.expression}')

  def __pow__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} ** {other!r}')

  def __rpow__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{other!r} ** {self.expression}')

  def __floordiv__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} // {other!r}')

  def __rfloordiv__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{other!r} // {self.expression}')

  def __truediv__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} / {other!r}')

  def __rtruediv__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{other!r} / {self.expression}')

  def __mod__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} % {other!r}')

  def __rmod__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{other!r} % {self.expression}')

  def __eq__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} == {other!r}')

  def __ne__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} != {other!r}')

  def __lt__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} < {other!r}')

  def __le__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} <= {other!r}')

  def __gt__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} > {other!r}')

  def __ge__(self, other: Any) -> '_UnresolvedExpression':
    return _UnresolvedExpression(f'{self.expression} >= {other!r}')

  def __iter__(self) -> Iterator['_UnresolvedExpression']:
    for i in range(3):
      yield getattr(self, f'item{i}')

  def keys(self) -> Iterator[Any]:
    for i in range(3):
      yield getattr(self, f'key{i}')

  def values(self) -> Iterator[Any]:
    for i in range(3):
      yield getattr(self, f'value{i}')

  def items(self) -> Iterator[tuple[Any, Any]]:
    for i in range(3):
      yield getattr(self, f'key{i}'), getattr(self, f'value{i}')



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from langfun.core.eval.v2 import example as example_lib
import pyglove as pg

Example = example_lib.Example


class ExampleTest(unittest.TestCase):

  def test_basic(self):
    error = pg.ErrorInfo(
        tag='ValueError',
        description='Bad input',
        stacktrace='...',
    )
    ex = Example(id=1, execution_status={
        'evaluate': pg.utils.TimeIt.Status(
            name='evaluation', elapse=1.0, error=error
        )
    })
    self.assertEqual(ex.error, error)
    self.assertFalse(ex.is_processed)
    self.assertTrue(ex.has_error)
    self.assertEqual(ex.elapse, 1.0)

    ex = Example(id=2, output=1)
    self.assertTrue(ex.is_processed)
    self.assertFalse(ex.has_error)
    self.assertIsNone(ex.elapse)

  def test_json_conversion(self):
    def input_func():
      class A(pg.Object):
        x: int

      class B(pg.Object):
        x: int = 1
        y: int = 2

      return [
          pg.Dict(
              a=A,
              b=B
          )
      ]

    inputs = input_func()
    ex = Example(
        id=1,
        input=inputs[0],
        output=inputs[0].a(1),
        metadata=dict(b=inputs[0].b())
    )
    # Serialize without input.
    json_str = pg.to_json_str(ex, exclude_input=True)
    self.assertEqual(
        pg.from_json_str(
            json_str,
            example_input_by_id=lambda i: inputs[i - 1],
            load_example_metadata=True,
        ),
        ex
    )
    self.assertEqual(
        pg.from_json_str(
            json_str,
            example_input_by_id=lambda i: inputs[i - 1],
            load_example_metadata=False,
        ),
        Example(
            id=1,
            input=inputs[0],
            output=inputs[0].a(1),
            metadata={}
        )
    )
    pg.JSONConvertible._TYPE_REGISTRY._type_to_cls_map.pop(
        inputs[0].a.__type_name__
    )
    pg.JSONConvertible._TYPE_REGISTRY._type_to_cls_map.pop(
        inputs[0].b.__type_name__
    )
    v = pg.from_json_str(json_str, auto_dict=True, load_example_metadata=True)
    v.output.pop('type_name')
    v.metadata.b.pop('type_name')
    self.assertEqual(
        v,
        Example(
            id=1,
            output=pg.Dict(x=1),
            metadata=dict(b=pg.Dict(x=1, y=2)),
        )
    )
    # Serialize with input.
    ex = Example(id=2, input=pg.Dict(x=1), output=pg.Dict(x=2))
    json_str = pg.to_json_str(ex, exclude_input=False)
    self.assertEqual(pg.from_json_str(json_str), ex)

  def test_html_view(self):
    ex = Example(
        id=1,
        input=pg.Dict(a=1, b=2),
        output=3,
        metadata=dict(sum=3),
        metric_metadata=dict(match=True),
    )
    self.assertNotIn(
        'next',
        ex.to_html(extra_flags=dict(num_examples=1)).content,
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import tempfile
import unittest

from langfun.core.eval.v2 import checkpointing
from langfun.core.eval.v2 import eval_test_helper
from langfun.core.eval.v2 import reporting
from langfun.core.eval.v2 import runners as runners_lib  # pylint: disable=unused-import
import pyglove as pg


class ReportingTest(unittest.TestCase):

  def test_reporting(self):
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_reporting')
    experiment = eval_test_helper.test_experiment()
    checkpointer = checkpointing.BulkCheckpointer('checkpoint.jsonl')
    reporter = reporting.HtmlReporter()
    run = experiment.run(root_dir, 'new', plugins=[checkpointer, reporter])
    self.assertTrue(
        pg.io.path_exists(os.path.join(run.output_root, 'summary.html'))
    )
    for leaf in experiment.leaf_nodes:
      self.assertTrue(
          pg.io.path_exists(run.output_path_for(leaf, 'index.html'))
      )
      for i in range(leaf.num_examples):
        self.assertTrue(
            pg.io.path_exists(run.output_path_for(leaf, f'{i + 1}.html'))
        )
      found_generation_log = False
      for log_entry in leaf._log_entries:
        if 'generated in' in log_entry.message:
          found_generation_log = True
          break
      self.assertTrue(found_generation_log)

    # Test warm start.
    root_dir = os.path.join(tempfile.mkdtemp(), 'test_reporting2')
    experiment = eval_test_helper.test_experiment()
    run = experiment.run(
        root_dir, 'new', plugins=[checkpointer, reporter],
        warm_start_from=run.output_root
    )
    self.assertTrue(
        pg.io.path_exists(os.path.join(run.output_root, 'summary.html'))
    )
    for leaf in experiment.leaf_nodes:
      self.assertTrue(
          pg.io.path_exists(run.output_path_for(leaf, 'index.html'))
      )
      for i in range(leaf.num_examples):
        self.assertTrue(
            pg.io.path_exists(run.output_path_for(leaf, f'{i + 1}.html'))
        )
      found_copy_log = False
      for log_entry in leaf._log_entries:
        if 'copied in' in log_entry.message:
          found_copy_log = True
          break
      self.assertTrue(found_copy_log)

  def test_index_html_generation_error(self):
    root_dir = os.path.join(
        tempfile.mkdtemp(),
        'test_reporting_with_index_html_generation_error'
    )
    experiment = (eval_test_helper
                  .test_experiment_with_index_html_generation_error())
    reporter = reporting.HtmlReporter()
    run = experiment.run(root_dir, 'new', plugins=[reporter])
    self.assertFalse(
        pg.io.path_exists(os.path.join(run.output_root, 'summary.html'))
    )
    for leaf in experiment.leaf_nodes:
      self.assertFalse(
          pg.io.path_exists(run.output_path_for(leaf, 'index.html'))
      )
    found_error_log = False
    for log_entry in experiment._log_entries:
      if log_entry.message.startswith('Failed to generate'):
        found_error_log = True
        break
    self.assertTrue(found_error_log)

  def test_example_html_generation_error(self):
    root_dir = os.path.join(
        tempfile.mkdtemp(),
        'test_reporting_with_example_html_generation_error'
    )
    experiment = (eval_test_helper
                  .test_experiment_with_example_html_generation_error())
    checkpointer = checkpointing.BulkCheckpointer('checkpoint.jsonl')
    reporter = reporting.HtmlReporter()
    run = experiment.run(root_dir, 'new', plugins=[checkpointer, reporter])
    self.assertTrue(
        pg.io.path_exists(os.path.join(run.output_root, 'summary.html'))
    )
    for leaf in experiment.leaf_nodes:
      self.assertTrue(
          pg.io.path_exists(run.output_path_for(leaf, 'index.html'))
      )
      for i in range(leaf.num_examples):
        self.assertFalse(
            pg.io.path_exists(run.output_path_for(leaf, f'{i + 1}.html'))
        )
    found_error_log = False
    for log_entry in experiment._log_entries:
      if log_entry.message.startswith('Failed to generate'):
        found_error_log = True
        break
    self.assertTrue(found_error_log)

    # Test warm start.
    root_dir = os.path.join(
        tempfile.mkdtemp(),
        'test_reporting_with_example_html_generation_error2'
    )
    experiment = (eval_test_helper
                  .test_experiment_with_example_html_generation_error())
    run = experiment.run(
        root_dir, 'new', plugins=[checkpointer, reporter],
        warm_start_from=run.output_root
    )
    self.assertTrue(
        pg.io.path_exists(os.path.join(run.output_root, 'summary.html'))
    )
    for leaf in experiment.leaf_nodes:
      self.assertTrue(
          pg.io.path_exists(run.output_path_for(leaf, 'index.html'))
      )
      for i in range(leaf.num_examples):
        self.assertFalse(
            pg.io.path_exists(run.output_path_for(leaf, f'{i + 1}.html'))
        )
    found_error_log = False
    for log_entry in experiment._log_entries:
      if log_entry.message.startswith('Skip copying'):
        found_error_log = True
        break
    self.assertTrue(found_error_log)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gemini API message conversion."""

import base64
from typing import Annotated, Any, Callable

import langfun.core as lf
from langfun.core import modalities as lf_modalities


class GeminiMessageConverter(lf.MessageConverter):
  """Converter to Gemini public API."""

  FORMAT_ID = 'gemini'

  chunk_preprocessor: Annotated[
      Callable[[str | lf.Modality], Any] | None,
      (
          'Chunk preprocessor for Langfun chunk to Gemini chunk conversion. '
          'It will be applied before each Langfun chunk is converted. '
          'If returns None, the chunk will be skipped.'
      ),
  ] = None

  def to_value(
      self, message: lf.Message, always_send_content: bool = True
  ) -> dict[str, Any]:
    """Converts a Langfun message to Gemini API."""
    parts = []
    for chunk in message.chunk():
      if self.chunk_preprocessor:
        chunk = self.chunk_preprocessor(chunk)
        if chunk is None:
          continue

      if isinstance(chunk, str):
        parts.append(self._convert_chunk(chunk, always_send_content))
      else:
        if isinstance(chunk, lf_modalities.Mime):
          modalities = [chunk]
        # NOTE(daiyip): preprocessing may convert a single chunk into
        # a list of chunks
        elif isinstance(chunk, list):
          modalities = chunk
        else:
          raise ValueError(f'Unsupported content type: {chunk!r}.')
        parts.extend(
            self._convert_chunk(c, always_send_content) for c in modalities
        )
    return dict(role=self.get_role(message), parts=parts)

  def _convert_chunk(
      self, chunk: str | lf.Modality, always_send_content: bool = True
  ) -> Any:
    """Converts a Langfun chunk to Gemini chunk."""
    if isinstance(chunk, str):
      return {'text': chunk}
    if not isinstance(chunk, lf_modalities.Mime):
      raise ValueError(f'Unsupported content chunk: {chunk!r}.')
    # NOTE(daiyip): special handling for YouTube video.
    if chunk.uri and chunk.uri.startswith('https://www.youtube.com/watch?v='):
      return {
          'fileData': {
              'mimeType': 'video/*',
              'fileUri': chunk.uri
          }
      }
    if chunk.is_text:
      return {'text': chunk.to_text()}
    if (
        not always_send_content
        and chunk.uri
        and chunk.uri.lower().startswith(('http:', 'https:', 'ftp:'))
    ):
      return {
          'fileData': {
              'mimeType': chunk.mime_type,
              'fileUri': chunk.uri,
          }
      }
    return {
        'inlineData': {
            'data': base64.b64encode(chunk.to_bytes()).decode(),
            'mimeType': chunk.mime_type,
        }
    }

  def from_value(self, value: dict[str, Any]) -> lf.Message:
    """Returns a Langfun message from Gemini message."""
    message_cls = self.get_message_cls(
        self._safe_read(value, 'role', default='model')
    )
    parts = self._safe_read(value, 'parts', default=[])
    assert isinstance(parts, list)

    chunks = []
    thought_chunks = []
    for part in parts:
      if 'text' in part:
        text = self._safe_read(part, 'text')
        if 'thought' in part:
          thought_chunks.append(text)
        else:
          chunks.append(text)
      elif 'thought' in part:
        # We occasionally encounter 'thought' without text.
        pass
      elif 'inlineData' in part:
        data = self._safe_read(part, 'inlineData')
        chunks.append(
            lf_modalities.Mime.class_from_mime_type(
                self._safe_read(data, 'mimeType')
            ).from_bytes(base64.b64decode(self._safe_read(data, 'data')))
        )
      elif 'fileData' in part:
        data = self._safe_read(part, 'fileData')
        chunks.append(
            lf_modalities.Mime.class_from_mime_type(
                self._safe_read(data, 'mimeType')
            ).from_uri(self._safe_read(data, 'fileUri'))
        )
      else:
        raise ValueError(f'Unsupported content part: {part!r}.')
    message = message_cls.from_chunks(chunks)
    if thought_chunks:
      message.set('thought', message_cls.from_chunks(thought_chunks))
    return message

  @classmethod
  def get_role(cls, message: lf.Message) -> str:
    """Returns the role of the message."""
    if isinstance(message, lf.AIMessage):
      return 'model'
    return super().get_role(message)

  @classmethod
  def get_message_cls(cls, role: str) -> type[lf.Message]:
    """Returns the message class of the message."""
    if role == 'model':
      return lf.AIMessage
    return super().get_message_cls(role)


def _as_gemini_format(
    self,
    chunk_preprocessor: Callable[[str | lf.Modality], Any] | None = None,
    **kwargs
) -> dict[str, Any]:
  """Returns a Gemini (REST) format message."""
  return GeminiMessageConverter(
      chunk_preprocessor=chunk_preprocessor, **kwargs
  ).to_value(self)


@classmethod
def _from_gemini_format(
    cls,
    gemini_message: dict[str, Any],
    **kwargs
) -> lf.Message:
  """Creates a Langfun message from the Gemini (REST) format message."""
  del cls
  return GeminiMessageConverter(**kwargs).from_value(gemini_message)

# Set shortcut methods in lf.Message.
lf.Message.as_gemini_format = _as_gemini_format
lf.Message.from_gemini_format = _from_gemini_format



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PDF tests."""

import unittest
from langfun.core.modalities import pdf as pdf_lib


pdf_bytes = (
    b'%PDF-1.1\n%\xc2\xa5\xc2\xb1\xc3\xab\n\n1 0 obj\n'
    b'<< /Type /Catalog\n     /Pages 2 0 R\n  >>\nendobj\n\n2 0 obj\n '
    b'<< /Type /Pages\n     /Kids [3 0 R]\n     '
    b'/Count 1\n     /MediaBox [0 0 300 144]\n  '
    b'>>\nendobj\n\n3 0 obj\n  '
    b'<<  /Type /Page\n      /Parent 2 0 R\n      /Resources\n       '
    b'<< /Font\n'
    b'<< /F1\n'
    b'<< /Type /Font\n'
    b'/Subtype /Type1\n'
    b'/BaseFont /Times-Roman\n'
    b'>>\n>>\n>>\n      '
    b'/Contents 4 0 R\n  >>\nendobj\n\n4 0 obj\n  '
    b'<< /Length 55 >>\nstream\n  BT\n    /F1 18 Tf\n    0 0 Td\n    '
    b'(Hello World) Tj\n  ET\nendstream\nendobj\n\nxref\n0 5\n0000000000 '
    b'65535 f \n0000000018 00000 n \n0000000077 00000 n \n0000000178 00000 n '
    b'\n0000000457 00000 n \ntrailer\n  <<  /Root 1 0 R\n      /Size 5\n  '
    b'>>\nstartxref\n565\n%%EOF\n'
)


class PdfTest(unittest.TestCase):

  def test_pdf(self):
    pdf = pdf_lib.PDF.from_bytes(pdf_bytes)
    self.assertEqual(pdf.mime_type, 'application/pdf')

  def test_repr_html(self):
    pdf = pdf_lib.PDF.from_bytes(pdf_bytes)
    self.assertIn(
        '<embed type="application/pdf" src="data:application/pdf;base64,',
        pdf._raw_html()
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Match-based evaluation."""

import abc
import io
import os
from typing import Any
import langfun.core as lf
from langfun.core.eval import base
import pyglove as pg


class Matching(base.Evaluation):
  """Base class for matching-based evaluation."""

  # CONSTANTS.
  MATCHES_JSON = 'matches.json'
  MISMATCHES_JSON = 'mismatches.json'

  MATCHES_HTML = 'matches.html'
  MISMATCHES_HTML = 'mismatches.html'

  @abc.abstractmethod
  def groundtruth(self, example: Any) -> Any:
    """Returns the groundtruth from an input example."""

  @abc.abstractmethod
  def answer(self, output: Any, example: Any) -> Any:
    """Returns the answer from the structure output."""

  @property
  def matches(self) -> list[tuple[int, Any, Any, lf.Message]]:
    """Returns the matches IDs, examples, outputs and the output messages."""
    return self._matches

  @property
  def num_matches(self) -> int:
    """Returns the number of matches between the answers and groundtruths."""
    return len(self.matches)

  @property
  def match_rate(self) -> float:
    if self.num_completed == 0:
      return 0.0
    return self.num_matches / self.num_completed

  @property
  def mismatches(self) -> list[tuple[int, Any, Any, lf.Message]]:
    """Returns the mismatches examples, outputs and output messages."""
    return self._mismatches

  @property
  def num_mismatches(self) -> int:
    """Returns the number of mismatches between the answers and groundtruths."""
    return len(self.mismatches)

  @property
  def mismatch_rate(self) -> float:
    if self.num_completed == 0:
      return 0.0
    return self.num_mismatches / self.num_completed

  @property
  def matches_link(self) -> str:
    """Returns the matches page."""
    return self.link(os.path.join(self.dir, Matching.MATCHES_HTML))

  @property
  def mismatches_link(self) -> str:
    return self.link(os.path.join(self.dir, Matching.MISMATCHES_HTML))

  def _reset(self) -> None:
    super()._reset()
    self._matches = []
    self._mismatches = []

  def audit_processed(
      self, example_idx: int, example: Any, output: Any, message: lf.Message,
      dryrun: bool = False
  ) -> None:
    groundtruth = self.groundtruth(example)
    answer = self.answer(output, example)

    if dryrun:
      lf.console.write('')
      lf.console.write(
          str(groundtruth),
          title='GROUDTRUTH',
          color='green',
      )
      lf.console.write('')
      lf.console.write(
          str(answer),
          title='ANSWER',
          color='blue',
      )

    if self.match(answer, groundtruth):
      self._matches.append((example_idx, example, output, message))
    else:
      self._mismatches.append((example_idx, example, output, message))

  def match(self, answer: Any, groundtruth: Any) -> bool:
    """Matches answer against the groundtruth. Subclasses can override."""
    return pg.eq(answer, groundtruth)

  def _eval_status(self, progress: lf.concurrent.Progress) -> dict[str, Any]:
    del progress
    return {
        'Matches': '%s (%d/%d)' % (
            self._format_rate(self.match_rate),
            self.num_matches,
            self.num_completed,
        ),
        'Mismatches': '%s (%d/%d)' % (
            self._format_rate(self.mismatch_rate),
            self.num_mismatches,
            self.num_completed,
        ),
        'Failed': '%s (%d/%d)' % (
            self._format_rate(self.failure_rate),
            self.num_failures,
            self.num_completed,
        ),
    }

  def _completion_status(self, run_status: str) -> str:
    assert self.result is not None
    m = self.result.metrics
    return (
        'COMPLETED(%s):'
        ' Matches=%s (%d/%d)'
        ' Mismatches=%s (%d/%d)'
        ' Failures=%s (%d/%d)'
    ) % (
        run_status,
        self._format_rate(m.match_rate),
        m.num_matches,
        m.total,
        self._format_rate(m.mismatch_rate),
        m.num_mismatches,
        m.total,
        self._format_rate(m.failure_rate),
        m.failures,
        m.total,
    )

  def finalize(self) -> pg.Dict:
    result = super().finalize()
    result.metrics.update(
        num_matches=self.num_matches,
        match_rate=self.match_rate,
        num_mismatches=self.num_mismatches,
        mismatch_rate=self.mismatch_rate,
    )
    return result

  def save(
      self, definition: bool = True, result: bool = True, report: bool = True
  ) -> None:
    super().save(definition, result, report)

    if report:
      pg.save(
          self._html([self._render_result, self._render_matches]),
          os.path.join(self.dir, Matching.MATCHES_HTML),
          file_format='txt',
      )
      pg.save(
          self._html([self._render_result, self._render_mismatches]),
          os.path.join(self.dir, Matching.MISMATCHES_HTML),
          file_format='txt',
      )

  def _render_result_header(self, s: io.StringIO):
    super()._render_result_header(s)
    s.write('<td>Mismatches</td>')
    s.write('<td>Matches</td>')

  def _render_result_row(self, s: io.StringIO):
    super()._render_result_row(s)
    s.write(
        '<td><span style="color:orange">%s</span>%s</td>'
        % (
            self._format_rate(self.mismatch_rate),
            '<a href="%s">(%d/%d)</a>'
            % (self.mismatches_link, self.num_mismatches, self.num_completed),
        )
    )
    s.write(
        '<td><span style="color:green">%s</span>%s</td>'
        % (
            self._format_rate(self.match_rate),
            '<a href="%s">(%d/%d)</a>'
            % (self.matches_link, self.num_matches, self.num_completed),
        )
    )

  def _render_summary_metrics(self, s: io.StringIO) -> None:
    """Renders metrics in HTML."""
    assert self.result is not None
    m = self.result.metrics
    self._render_link(
        s,
        'Matches (%d/%d)' % (m.num_matches, m.total),
        self._format_rate(m.match_rate),
        'color:green',
        lambda: self.matches_link,
    )
    s.write(' | ')
    self._render_link(
        s,
        'Mismatches (%d/%d)' % (m.num_mismatches, m.total),
        self._format_rate(m.mismatch_rate),
        'color:orange',
        lambda: self.mismatches_link,
    )
    s.write(' | ')
    super()._render_summary_metrics(s)

  def _render_matches(self, s: io.StringIO) -> None:
    """Formats the matched cases into html."""
    s.write('<h2> Matches (Correct) </h2>')
    s.write('<div style="white-space:pre">\n')
    s.write(
        '<table style="border: 1px solid;">'
        '<tr class="header">'
        '<td>No.</td><td>Input</td><td>Output</td>'
        '<td>Prompt/Response Chain</td>'
        '</tr>'
    )
    def _maybe_html(v, root_indent: int):
      del root_indent
      if hasattr(v, '_repr_html_'):
        return v._repr_html_()  # pylint: disable=protected-access
      # Fall back to the default format.
      return None

    for i, (_, example, output, message) in enumerate(self.matches):
      bgcolor = 'white' if i % 2 == 0 else '#DDDDDD'
      s.write(f'<tr style="background-color: {bgcolor}"><td>{i + 1}</td>')
      input_str = pg.Html.escape(
          pg.format(
              example, verbose=False, max_bytes_len=32,
              custom_format=_maybe_html
          )
      )
      s.write(f'<td style="color:green;white-space:pre-wrap">{input_str}</td>')
      output_str = pg.Html.escape(
          pg.format(
              output, verbose=False, max_bytes_len=32,
              custom_format=_maybe_html
          )
      )
      s.write(f'<td style="color:blue;white-space:pre-wrap">{output_str}</td>')
      s.write('<td>')
      self._render_message(message, s)
      s.write('</td>')
      s.write('</tr>')
    s.write('</table></div>')

  def _render_mismatches(self, s: io.StringIO) -> None:
    """Formats the mismatched cases into html."""
    s.write('<h2> Mismatches (Incorrect) </h2>')
    s.write('<div style="white-space:pre">\n')
    s.write(
        '<table style="border: 1px solid;">'
        '<tr class="header">'
        '<td>No.</td><td>Input</td><td>Output</td>'
        '<td>Prompt/Response Chain</td>'
        '</tr>'
    )

    for i, (_, example, output, message) in enumerate(self.mismatches):
      bgcolor = 'white' if i % 2 == 0 else '#DDDDDD'
      s.write(f'<tr style="background-color: {bgcolor}"><td>{i + 1}</td>')
      input_str = pg.format(example, verbose=False, max_bytes_len=32)
      s.write(f'<td style="color:green;white-space:pre-wrap">{input_str}</td>')
      output_str = pg.format(output, verbose=False, max_bytes_len=32)
      s.write(
          f'<td style="color:magenta;white-space:pre-wrap">{output_str}</td>'
      )
      s.write('<td>')
      self._render_message(message, s)
      s.write('</td>')
      s.write('</tr>')
    s.write('</table></div>')



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for VertexAI models."""

import os
import unittest
from unittest import mock

from google import auth
from google.auth import exceptions
import langfun.core as lf
from langfun.core.llms import rest
from langfun.core.llms import vertexai
import pyglove as pg


class VertexAITest(unittest.TestCase):
  """Tests for Vertex model with REST API."""

  @mock.patch.object(vertexai.VertexAI, 'credentials', new=True)
  def test_project_and_location_check(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `project`'):
      _ = vertexai.VertexAIGemini15Pro()._api_initialized

    with self.assertRaisesRegex(ValueError, 'Please specify `location`'):
      _ = vertexai.VertexAIGemini15Pro(
          project='abc', location=None)._api_initialized

    self.assertTrue(
        vertexai.VertexAIGemini15Pro(
            project='abc', location='us-central1'
        )._api_initialized
    )

    os.environ['VERTEXAI_PROJECT'] = 'abc'
    os.environ['VERTEXAI_LOCATION'] = 'us-central1'
    model = vertexai.VertexAIGemini15Pro(location=pg.MISSING_VALUE)
    self.assertEqual(model.resource_id, 'vertexai://gemini-1.5-pro-002')
    self.assertIn('us-central1', model.api_endpoint)
    self.assertTrue(model._api_initialized)
    self.assertIsNotNone(model.session())
    del os.environ['VERTEXAI_PROJECT']
    del os.environ['VERTEXAI_LOCATION']

  def test_auth_refresh_error(self):
    def _auth_refresh_error(*args, **kwargs):
      del args, kwargs
      raise exceptions.RefreshError('Cannot refresh token')

    with self.assertRaisesRegex(
        lf.concurrent.RetryError,
        'Failed to refresh Google authentication credentials'
    ):
      with (
          mock.patch.object(auth, 'default') as mock_auth,
          mock.patch.object(rest.REST, '_sample_single') as mock_sample_single
      ):
        mock_auth.return_value = mock.MagicMock(), None
        mock_sample_single.side_effect = _auth_refresh_error
        model = vertexai.VertexAIGemini15Pro(
            project='abc', location='us-central1', max_attempts=1
        )
        model('hi')


class VertexAIAnthropicTest(unittest.TestCase):
  """Tests for VertexAI Anthropic models."""

  def test_basics(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `project`'):
      lm = vertexai.VertexAIClaude35Sonnet_20241022()
      lm('hi')

    model = vertexai.VertexAIClaude35Sonnet_20241022(project='langfun')
    self.assertEqual(model.resource_id, 'vertexai://claude-3-5-sonnet-20241022')
    # Map a Anthropic model back to VertexAI model.
    self.assertTrue(
        vertexai.VertexAIAnthropic(
            'claude-3-5-sonnet-20241022', project='langfun'
        ).model,
        'claude-3-5-sonnet-v2@20241022',
    )
    # NOTE(daiyip): For OSS users, default credentials are not available unless
    # users have already set up their GCP project. Therefore we ignore the
    # exception here.
    try:
      model._initialize()
    except exceptions.DefaultCredentialsError:
      pass

    self.assertEqual(
        model.api_endpoint,
        (
            'https://us-east5-aiplatform.googleapis.com/v1/projects/'
            'langfun/locations/us-east5/publishers/anthropic/'
            'models/claude-3-5-sonnet-v2@20241022:streamRawPredict'
        )
    )
    self.assertEqual(
        model.headers,
        {
            'Content-Type': 'application/json; charset=utf-8',
        },
    )
    request = model.request(
        lf.UserMessage('hi'), lf.LMSamplingOptions(temperature=0.0),
    )
    self.assertEqual(
        request,
        {
            'anthropic_version': 'vertex-2023-10-16',
            'max_tokens': 8192,
            'messages': [
                {'content': [{'text': 'hi', 'type': 'text'}], 'role': 'user'}
            ],
            'stream': False,
            'temperature': 0.0,
            'top_k': 40,
        },
    )

  def test_lm_get(self):
    self.assertIsInstance(
        lf.LanguageModel.get('gemini-2.0-flash'),
        vertexai.VertexAIGemini,
    )
    self.assertIsInstance(
        lf.LanguageModel.get('claude-3-5-sonnet-v2@20241022'),
        vertexai.VertexAIAnthropic,
    )
    self.assertIsInstance(
        lf.LanguageModel.get('llama-3.1-405b-instruct-maas'),
        vertexai.VertexAILlama,
    )
    self.assertIsInstance(
        lf.LanguageModel.get('mistral-large-2411'),
        vertexai.VertexAIMistral,
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for groundtruth matching."""

import os
import tempfile
from typing import Any
import unittest

import langfun.core as lf
from langfun.core.eval import base
from langfun.core.eval import matching
from langfun.core.llms import fake
import pyglove as pg


# We put class definitions outside the functors just to make it easier
# to refer to them in test.


class Solution(pg.Object):
  final_answer: int


class SolutionForCompletion(pg.Object):
  question: str
  final_answer: int


@pg.functor
def answer_schema():
  return Solution


@pg.functor
def complete_schema():
  return SolutionForCompletion


class MyTask(matching.Matching):

  def groundtruth(self, example: Any) -> Any:
    return example.groundtruth

  def answer(self, output: Any, example: Any) -> Any:
    return output.final_answer


def eval_set(
    eval_id: str,
    method: str,
    schema_fn,
    lm: lf.LanguageModel,
    use_cache: bool = True,
):
  """Creates an evaluation object for testing."""
  return MyTask(
      root_dir=os.path.join(tempfile.gettempdir(), eval_id),
      inputs=base.as_inputs([
          pg.Dict(question='Compute 1 + 1', groundtruth=2),
          pg.Dict(question='Compute 1 + 2', groundtruth=3),
          pg.Dict(question='Compute 1 + 3', groundtruth=4),
          pg.Dict(question='Compute 1 + 1', groundtruth=2),
      ]),
      method=method,
      prompt='{{example.question}}',
      schema_fn=schema_fn,
      lm=lm,
      use_cache=use_cache,
      max_workers=1,
  )


class MatchingTest(unittest.TestCase):
  """Matching test."""

  def setUp(self):
    super().setUp()
    pg.symbolic.set_save_handler(pg.symbolic.default_save_handler)
    pg.symbolic.set_load_handler(pg.symbolic.default_load_handler)

  def test_run(self):
    lm = fake.StaticSequence([
        'Solution(final_answer=2)',
        '3',
        'Solution(final_answer=3)',
    ])

    s = eval_set('match_run_test', 'query', schema_fn=answer_schema(), lm=lm)
    s.run()
    result_without_id = s.result.copy()
    result_without_id['experiment_setup'].pop('id')
    self.assertEqual(
        result_without_id,
        dict(
            experiment_setup=dict(
                dir=s.dir,
                model='StaticSequence',
                prompt_template='{{example.question}}',
                method='query',
                schema_fn='answer_schema()',
            ),
            cache_stats=dict(
                use_cache=True,
                num_queries=4,
                num_hits=1,
                num_updates=3,
            ),
            metrics=dict(
                total=4,
                failures=1,
                failure_rate=0.25,
                oop_failures=1,
                oop_failure_rate=0.25,
                non_oop_failures=0,
                non_oop_failure_rate=0.0,
                failure_breakdown={
                    'MappingError.SchemaError.TypeError': 1
                },
                num_matches=2,
                match_rate=0.5,
                num_mismatches=1,
                mismatch_rate=0.25,
            ),
            usage=s.result.usage,
        ),
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, matching.Matching.EXPERIMENT_JSON
            )
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, matching.Matching.RESULT_JSON)
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, matching.Matching.CACHE_JSON)
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, matching.Matching.OOP_FAILURES_JSON
            )
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, matching.Matching.NON_OOP_FAILURES_JSON
            )
        )
    )
    self.assertTrue(
        os.path.exists(os.path.join(s.root_dir, matching.Matching.SUMMARY_HTML))
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, matching.Matching.INDEX_HTML)
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, matching.Matching.OOP_FAILURES_HTML
            )
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, matching.Matching.NON_OOP_FAILURES_HTML
            )
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(s.dir, matching.Matching.MATCHES_HTML)
        )
    )
    self.assertTrue(
        os.path.exists(
            os.path.join(
                s.dir, matching.Matching.MISMATCHES_HTML
            )
        )
    )


if __name__ == '__main__':
  unittest.main()
